########################################################################
# File name: xso.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import aioxmpp.forms
import aioxmpp.stanza
import aioxmpp.xso as xso

from enum import Enum

from aioxmpp.utils import namespaces


class Feature(Enum):
    ACCESS_AUTHORIZE = \
        "http://jabber.org/protocol/pubsub#access-authorize"
    ACCESS_OPEN = \
        "http://jabber.org/protocol/pubsub#access-open"
    ACCESS_PRESENCE = \
        "http://jabber.org/protocol/pubsub#access-presence"
    ACCESS_ROSTER = \
        "http://jabber.org/protocol/pubsub#access-roster"
    ACCESS_WHITELIST = \
        "http://jabber.org/protocol/pubsub#access-whitelist"
    AUTO_CREATE = \
        "http://jabber.org/protocol/pubsub#auto-create"
    AUTO_SUBSCRIBE = \
        "http://jabber.org/protocol/pubsub#auto-subscribe"
    COLLECTIONS = \
        "http://jabber.org/protocol/pubsub#collections"
    CONFIG_NODE = \
        "http://jabber.org/protocol/pubsub#config-node"
    CREATE_AND_CONFIGURE = \
        "http://jabber.org/protocol/pubsub#create-and-configure"
    CREATE_NODES = \
        "http://jabber.org/protocol/pubsub#create-nodes"
    DELETE_ITEMS = \
        "http://jabber.org/protocol/pubsub#delete-items"
    DELETE_NODES = \
        "http://jabber.org/protocol/pubsub#delete-nodes"
    FILTERED_NOTIFICATIONS = \
        "http://jabber.org/protocol/pubsub#filtered-notifications"
    GET_PENDING = \
        "http://jabber.org/protocol/pubsub#get-pending"
    INSTANT_NODES = \
        "http://jabber.org/protocol/pubsub#instant-nodes"
    ITEM_IDS = \
        "http://jabber.org/protocol/pubsub#item-ids"
    LAST_PUBLISHED = \
        "http://jabber.org/protocol/pubsub#last-published"
    LEASED_SUBSCRIPTION = \
        "http://jabber.org/protocol/pubsub#leased-subscription"
    MANAGE_SUBSCRIPTIONS = \
        "http://jabber.org/protocol/pubsub#manage-subscriptions"
    MEMBER_AFFILIATION = \
        "http://jabber.org/protocol/pubsub#member-affiliation"
    META_DATA = \
        "http://jabber.org/protocol/pubsub#meta-data"
    MODIFY_AFFILIATIONS = \
        "http://jabber.org/protocol/pubsub#modify-affiliations"
    MULTI_COLLECTION = \
        "http://jabber.org/protocol/pubsub#multi-collection"
    MULTI_SUBSCRIBE = \
        "http://jabber.org/protocol/pubsub#multi-subscribe"
    OUTCAST_AFFILIATION = \
        "http://jabber.org/protocol/pubsub#outcast-affiliation"
    PERSISTENT_ITEMS = \
        "http://jabber.org/protocol/pubsub#persistent-items"
    PRESENCE_NOTIFICATIONS = \
        "http://jabber.org/protocol/pubsub#presence-notifications"
    PRESENCE_SUBSCRIBE = \
        "http://jabber.org/protocol/pubsub#presence-subscribe"
    PUBLISH = \
        "http://jabber.org/protocol/pubsub#publish"
    PUBLISH_OPTIONS = \
        "http://jabber.org/protocol/pubsub#publish-options"
    PUBLISH_ONLY_AFFILIATION = \
        "http://jabber.org/protocol/pubsub#publish-only-affiliation"
    PUBLISHER_AFFILIATION = \
        "http://jabber.org/protocol/pubsub#publisher-affiliation"
    PURGE_NODES = \
        "http://jabber.org/protocol/pubsub#purge-nodes"
    RETRACT_ITEMS = \
        "http://jabber.org/protocol/pubsub#retract-items"
    RETRIEVE_AFFILIATIONS = \
        "http://jabber.org/protocol/pubsub#retrieve-affiliations"
    RETRIEVE_DEFAULT = \
        "http://jabber.org/protocol/pubsub#retrieve-default"
    RETRIEVE_ITEMS = \
        "http://jabber.org/protocol/pubsub#retrieve-items"
    RETRIEVE_SUBSCRIPTIONS = \
        "http://jabber.org/protocol/pubsub#retrieve-subscriptions"
    SUBSCRIBE = \
        "http://jabber.org/protocol/pubsub#subscribe"
    SUBSCRIPTION_OPTIONS = \
        "http://jabber.org/protocol/pubsub#subscription-options"
    SUBSCRIPTION_NOTIFICATIONS = \
        "http://jabber.org/protocol/pubsub#subscription-notifications"


namespaces.xep0060_features = Feature
namespaces.xep0060 = "http://jabber.org/protocol/pubsub"
namespaces.xep0060_errors = "http://jabber.org/protocol/pubsub#errors"
namespaces.xep0060_event = "http://jabber.org/protocol/pubsub#event"
namespaces.xep0060_owner = "http://jabber.org/protocol/pubsub#owner"


class Affiliation(xso.XSO):
    TAG = (namespaces.xep0060, "affiliation")

    node = xso.Attr(
        "node",
        default=None
    )

    affiliation = xso.Attr(
        "affiliation",
        validator=xso.RestrictToSet({
            "member",
            "none",
            "outcast",
            "owner",
            "publisher",
            "publish-only",
        }),
    )

    def __init__(self, affiliation, node=None):
        super().__init__()
        self.affiliation = affiliation
        self.node = node


class Affiliations(xso.XSO):
    TAG = (namespaces.xep0060, "affiliations")

    node = xso.Attr(
        "node",
        default=None
    )

    affiliations = xso.ChildList(
        [Affiliation],
    )

    def __init__(self, affiliations=[], node=None):
        super().__init__()
        self.affiliations[:] = affiliations
        self.node = node


class Configure(xso.XSO):
    TAG = (namespaces.xep0060, "configure")

    data = xso.Child([
        aioxmpp.forms.Data,
    ])


class Create(xso.XSO):
    TAG = (namespaces.xep0060, "create")

    node = xso.Attr(
        "node",
        default=None
    )


class Default(xso.XSO):
    TAG = (namespaces.xep0060, "default")

    node = xso.Attr(
        "node",
        default=None
    )

    type_ = xso.Attr(
        "type",
        validator=xso.RestrictToSet({
            "leaf",
            "collection",
        }),
        default="leaf",
    )

    data = xso.Child([
        aioxmpp.forms.Data,
    ])

    def __init__(self, *, node=None, data=None):
        super().__init__()
        self.node = node
        self.data = data


class Item(xso.XSO):
    TAG = (namespaces.xep0060, "item")

    id_ = xso.Attr(
        "id",
        default=None
    )

    registered_payload = xso.Child([], strict=True)

    unregistered_payload = xso.Collector()

    def __init__(self, id_=None):
        super().__init__()
        self.id_ = id_


class Items(xso.XSO):
    TAG = (namespaces.xep0060, "items")

    max_items = xso.Attr(
        (None, "max_items"),
        type_=xso.Integer(),
        validator=xso.NumericRange(min_=1),
        default=None,
    )

    node = xso.Attr(
        "node",
    )

    subid = xso.Attr(
        "subid",
        default=None
    )

    items = xso.ChildList(
        [Item]
    )

    def __init__(self, node, subid=None, max_items=None):
        super().__init__()
        self.node = node
        self.subid = subid
        self.max_items = max_items


class Options(xso.XSO):
    TAG = (namespaces.xep0060, "options")

    jid = xso.Attr(
        "jid",
        type_=xso.JID()
    )

    node = xso.Attr(
        "node",
        default=None
    )

    subid = xso.Attr(
        "subid",
        default=None
    )

    data = xso.Child([
        aioxmpp.forms.Data,
    ])

    def __init__(self, jid, node=None, subid=None):
        super().__init__()
        self.jid = jid
        self.node = node
        self.subid = subid


class Publish(xso.XSO):
    TAG = (namespaces.xep0060, "publish")

    node = xso.Attr(
        "node",
    )

    item = xso.Child([
        Item
    ])


class Retract(xso.XSO):
    TAG = (namespaces.xep0060, "retract")

    node = xso.Attr(
        "node",
    )

    item = xso.Child([
        Item
    ])

    notify = xso.Attr(
        "notify",
        type_=xso.Bool(),
        default=False,
    )


class Subscribe(xso.XSO):
    TAG = (namespaces.xep0060, "subscribe")

    jid = xso.Attr(
        "jid",
        type_=xso.JID()
    )

    node = xso.Attr(
        "node",
        default=None
    )

    def __init__(self, jid, node=None):
        super().__init__()
        self.jid = jid
        self.node = node


class SubscribeOptions(xso.XSO):
    TAG = (namespaces.xep0060, "subscribe-options")

    required = xso.ChildTag(
        [
            (namespaces.xep0060, "required"),
        ],
        allow_none=True
    )


class Subscription(xso.XSO):
    TAG = (namespaces.xep0060, "subscription")

    jid = xso.Attr(
        "jid",
        type_=xso.JID()
    )

    node = xso.Attr(
        "node",
        default=None
    )

    subid = xso.Attr(
        "subid",
        default=None
    )

    subscription = xso.Attr(
        "subscription",
        validator=xso.RestrictToSet({
            "none",
            "pending",
            "subscribed",
            "unconfigured",
        }),
        default=None
    )

    subscribe_options = xso.Child(
        [SubscribeOptions]
    )

    def __init__(self, jid, node=None, subid=None, *, subscription=None):
        super().__init__()
        self.jid = jid
        self.node = node
        self.subid = subid
        self.subscription = subscription


class Subscriptions(xso.XSO):
    TAG = (namespaces.xep0060, "subscriptions")

    node = xso.Attr(
        "node",
        default=None
    )

    subscriptions = xso.ChildList(
        [Subscription],
    )

    def __init__(self, *, subscriptions=[], node=None):
        super().__init__()
        self.node = node
        self.subscriptions[:] = subscriptions


class Unsubscribe(xso.XSO):
    TAG = (namespaces.xep0060, "unsubscribe")

    jid = xso.Attr(
        "jid",
        type_=xso.JID()
    )

    node = xso.Attr(
        "node",
        default=None
    )

    subid = xso.Attr(
        "subid",
        default=None
    )

    def __init__(self, jid, node=None, subid=None):
        super().__init__()
        self.jid = jid
        self.node = node
        self.subid = subid


@aioxmpp.stanza.IQ.as_payload_class
class Request(xso.XSO):
    """
    This XSO represents the ``<pubsub/>`` IQ payload from the generic pubsub
    namespace (``http://jabber.org/protocol/pubsub``). It can carry different
    types of payload.

    .. attribute:: payload

       This is the generic payload attribute. It supports the following
       classes:

       .. autosummary::

          Affiliations
          Create
          Default
          Items
          Publish
          Retract
          Subscribe
          Subscription
          Subscriptions
          Unsubscribe

    .. attribute:: options

       As :class:`Options` may both be used independently and along with
       another :attr:`payload`, they have their own attribute.

       Independent of their use, :class:`Options` objects are always available
       here. If they are used without another payload, the :attr:`payload`
       attribute is :data:`None`.

    .. attribute:: configure

       As :class:`Configure` may both be used independently and along with
       another :attr:`payload`, it has its own attribute.

       Independent of their use, :class:`Configure` objects are always
       available here. If they are used without another payload, the
       :attr:`payload` attribute is :data:`None`.

    """
    TAG = (namespaces.xep0060, "pubsub")

    payload = xso.Child([
        Affiliations,
        Create,
        Default,
        Items,
        Publish,
        Retract,
        Subscribe,
        Subscription,
        Subscriptions,
        Unsubscribe,
    ])

    options = xso.Child([
        Options,
    ])

    configure = xso.Child([
        Configure,
    ])

    def __init__(self, payload=None):
        super().__init__()
        self.payload = payload


aioxmpp.stanza.Message.xep0060_request = xso.Child([
    Request
])


class EventAssociate(xso.XSO):
    TAG = (namespaces.xep0060_event, "associate")

    node = xso.Attr(
        "node",
    )


class EventDisassociate(xso.XSO):
    TAG = (namespaces.xep0060_event, "disassociate")

    node = xso.Attr(
        "node",
    )


class EventCollection(xso.XSO):
    TAG = (namespaces.xep0060_event, "collection")

    assoc = xso.Child([
        EventAssociate,
        EventDisassociate,
    ])


class EventRedirect(xso.XSO):
    TAG = (namespaces.xep0060_event, "redirect")

    uri = xso.Attr(
        "uri",
    )

    def __init__(self, uri):
        super().__init__()
        self.uri = uri


class EventDelete(xso.XSO):
    TAG = (namespaces.xep0060_event, "delete")

    _redirect = xso.Child([
        EventRedirect,
    ])

    node = xso.Attr(
        "node",
    )

    def __init__(self, node, *, redirect_uri=None):
        super().__init__()
        self.node = node
        if redirect_uri is not None:
            self._redirect = EventRedirect(redirect_uri)

    @property
    def redirect_uri(self):
        if self._redirect is None:
            return None
        return self._redirect.uri

    @redirect_uri.setter
    def redirect_uri(self, value):
        if value is None:
            del self._redirect
            return
        self._redirect = EventRedirect(value)

    @redirect_uri.deleter
    def redirect_uri(self):
        del self._redirect


class EventRetract(xso.XSO):
    TAG = (namespaces.xep0060_event, "retract")

    id_ = xso.Attr(
        "id",
    )

    def __init__(self, id_):
        super().__init__()
        self.id_ = id_


class EventItem(xso.XSO):
    TAG = (namespaces.xep0060_event, "item")

    id_ = xso.Attr(
        "id",
        default=None,
    )

    node = xso.Attr(
        "node",
        default=None,
    )

    publisher = xso.Attr(
        "publisher",
        default=None,
    )

    registered_payload = xso.Child([], strict=True)

    unregistered_payload = xso.Collector()

    def __init__(self, payload, *, id_=None):
        super().__init__()
        self.registered_payload = payload
        self.id_ = id_


class EventItems(xso.XSO):
    TAG = (namespaces.xep0060_event, "items")

    node = xso.Attr(
        "node",
    )

    retracts = xso.ChildList([EventRetract])

    items = xso.ChildList([EventItem])

    def __init__(self, node, *, items=[], retracts=[]):
        super().__init__()
        self.items[:] = items
        self.retracts[:] = retracts
        self.node = node


class EventPurge(xso.XSO):
    TAG = (namespaces.xep0060_event, "purge")

    node = xso.Attr(
        "node",
    )


class EventSubscription(xso.XSO):
    TAG = (namespaces.xep0060_event, "subscription")

    jid = xso.Attr(
        "jid",
        type_=xso.JID()
    )

    node = xso.Attr(
        "node",
        default=None
    )

    subid = xso.Attr(
        "subid",
        default=None
    )

    subscription = xso.Attr(
        "subscription",
        validator=xso.RestrictToSet({
            "none",
            "pending",
            "subscribed",
            "unconfigured",
        }),
        default=None
    )

    expiry = xso.Attr(
        "expiry",
        type_=xso.DateTime(),
    )


class EventConfiguration(xso.XSO):
    TAG = (namespaces.xep0060_event, "configuration")

    node = xso.Attr(
        "node",
        default=None
    )

    data = xso.Child([
        aioxmpp.forms.Data,
    ])


class Event(xso.XSO):
    TAG = (namespaces.xep0060_event, "event")

    payload = xso.Child([
        EventCollection,
        EventConfiguration,
        EventDelete,
        EventItems,
        EventPurge,
        EventSubscription,
    ])

    def __init__(self, payload=None):
        super().__init__()
        self.payload = payload


aioxmpp.stanza.Message.xep0060_event = xso.Child([
    Event
])


class OwnerAffiliation(xso.XSO):
    TAG = (namespaces.xep0060_owner, "affiliation")

    affiliation = xso.Attr(
        "affiliation",
        validator=xso.RestrictToSet({
            "member",
            "outcast",
            "owner",
            "publisher",
            "publish-only",
            "none",
        }),
        validate=xso.ValidateMode.ALWAYS,
    )

    jid = xso.Attr(
        "jid",
        type_=xso.JID(),
    )

    def __init__(self, jid, affiliation):
        super().__init__()
        self.jid = jid
        self.affiliation = affiliation


class OwnerAffiliations(xso.XSO):
    TAG = (namespaces.xep0060_owner, "affiliations")

    affiliations = xso.ChildList([OwnerAffiliation])

    node = xso.Attr(
        "node",
    )

    def __init__(self, node, *, affiliations=[]):
        super().__init__()
        self.node = node
        self.affiliations[:] = affiliations


class OwnerConfigure(xso.XSO):
    TAG = (namespaces.xep0060_owner, "configure")

    node = xso.Attr(
        "node",
        default=None,
    )

    data = xso.Child([
        aioxmpp.forms.Data,
    ])


class OwnerDefault(xso.XSO):
    TAG = (namespaces.xep0060_owner, "default")

    data = xso.Child([
        aioxmpp.forms.Data,
    ])


class OwnerRedirect(xso.XSO):
    TAG = (namespaces.xep0060_owner, "redirect")

    uri = xso.Attr(
        "uri",
    )

    def __init__(self, uri):
        super().__init__()
        self.uri = uri


class OwnerDelete(xso.XSO):
    TAG = (namespaces.xep0060_owner, "delete")

    _redirect = xso.Child([OwnerRedirect])

    node = xso.Attr(
        "node",
    )

    def __init__(self, node, *, redirect_uri=None):
        super().__init__()
        self.node = node
        if redirect_uri is not None:
            self._redirect = OwnerRedirect(redirect_uri)

    @property
    def redirect_uri(self):
        if self._redirect is None:
            return None
        return self._redirect.uri

    @redirect_uri.setter
    def redirect_uri(self, value):
        if value is None:
            del self._redirect
            return
        self._redirect = OwnerRedirect(value)

    @redirect_uri.deleter
    def redirect_uri(self):
        del self._redirect


class OwnerPurge(xso.XSO):
    TAG = (namespaces.xep0060_owner, "purge")

    node = xso.Attr(
        "node",
    )

    def __init__(self, node):
        super().__init__()
        self.node = node


class OwnerSubscription(xso.XSO):
    TAG = (namespaces.xep0060_owner, "subscription")

    subscription = xso.Attr(
        "subscription",
        validator=xso.RestrictToSet({
            "none",
            "pending",
            "subscribed",
            "unconfigured",
        }),
        validate=xso.ValidateMode.ALWAYS
    )

    jid = xso.Attr(
        "jid",
        type_=xso.JID(),
    )

    def __init__(self, jid, subscription):
        super().__init__()
        self.jid = jid
        self.subscription = subscription


class OwnerSubscriptions(xso.XSO):
    TAG = (namespaces.xep0060_owner, "subscriptions")

    node = xso.Attr(
        "node",
    )

    subscriptions = xso.ChildList([OwnerSubscription])

    def __init__(self, node, *, subscriptions=[]):
        super().__init__()
        self.node = node
        self.subscriptions[:] = subscriptions


@aioxmpp.stanza.IQ.as_payload_class
class OwnerRequest(xso.XSO):
    TAG = (namespaces.xep0060_owner, "pubsub")

    payload = xso.Child([
        OwnerAffiliations,
        OwnerConfigure,
        OwnerDefault,
        OwnerDelete,
        OwnerPurge,
        OwnerSubscriptions,
    ])

    def __init__(self, payload):
        super().__init__()
        self.payload = payload


def as_payload_class(cls):
    """
    Register the given class `cls` as Publish-Subscribe payload on both
    :class:`Item` and :class:`EventItem`.

    Return the class, to allow this to be used as decorator.
    """

    Item.register_child(
        Item.registered_payload,
        cls,
    )

    EventItem.register_child(
        EventItem.registered_payload,
        cls,
    )

    return cls


ClosedNode = aioxmpp.stanza.make_application_error(
    "ClosedNode",
    (namespaces.xep0060_errors, "closed-node"),
)

ConfigurationRequired = aioxmpp.stanza.make_application_error(
    "ConfigurationRequired",
    (namespaces.xep0060_errors, "configuration-required"),
)

InvalidJID = aioxmpp.stanza.make_application_error(
    "InvalidJID",
    (namespaces.xep0060_errors, "invalid-jid"),
)

InvalidOptions = aioxmpp.stanza.make_application_error(
    "InvalidOptions",
    (namespaces.xep0060_errors, "invalid-options"),
)

InvalidPayload = aioxmpp.stanza.make_application_error(
    "InvalidPayload",
    (namespaces.xep0060_errors, "invalid-payload"),
)

InvalidSubID = aioxmpp.stanza.make_application_error(
    "InvalidSubID",
    (namespaces.xep0060_errors, "invalid-subid"),
)

ItemForbidden = aioxmpp.stanza.make_application_error(
    "ItemForbidden",
    (namespaces.xep0060_errors, "item-forbidden"),
)

ItemRequired = aioxmpp.stanza.make_application_error(
    "ItemRequired",
    (namespaces.xep0060_errors, "item-required"),
)

JIDRequired = aioxmpp.stanza.make_application_error(
    "JIDRequired",
    (namespaces.xep0060_errors, "jid-required"),
)

MaxItemsExceeded = aioxmpp.stanza.make_application_error(
    "MaxItemsExceeded",
    (namespaces.xep0060_errors, "max-items-exceeded"),
)

MaxNodesExceeded = aioxmpp.stanza.make_application_error(
    "MaxNodesExceeded",
    (namespaces.xep0060_errors, "max-nodes-exceeded"),
)

NodeIDRequired = aioxmpp.stanza.make_application_error(
    "NodeIDRequired",
    (namespaces.xep0060_errors, "nodeid-required"),
)

NotInRosterGroup = aioxmpp.stanza.make_application_error(
    "NotInRosterGroup",
    (namespaces.xep0060_errors, "not-in-roster-group"),
)

NotSubscribed = aioxmpp.stanza.make_application_error(
    "NotSubscribed",
    (namespaces.xep0060_errors, "not-subscribed"),
)

PayloadTooBig = aioxmpp.stanza.make_application_error(
    "PayloadTooBig",
    (namespaces.xep0060_errors, "payload-too-big"),
)

PayloadRequired = aioxmpp.stanza.make_application_error(
    "PayloadRequired",
    (namespaces.xep0060_errors, "payload-required"),
)

PendingSubscription = aioxmpp.stanza.make_application_error(
    "PendingSubscription",
    (namespaces.xep0060_errors, "pending-subscription"),
)

PresenceSubscriptionRequired = aioxmpp.stanza.make_application_error(
    "PresenceSubscriptionRequired",
    (namespaces.xep0060_errors, "presence-subscription-required"),
)

SubIDRequired = aioxmpp.stanza.make_application_error(
    "SubIDRequired",
    (namespaces.xep0060_errors, "subid-required"),
)

TooManySubscriptions = aioxmpp.stanza.make_application_error(
    "TooManySubscriptions",
    (namespaces.xep0060_errors, "too-many-subscriptions"),
)


@aioxmpp.stanza.Error.as_application_condition
class Unsupported(xso.XSO):
    TAG = (namespaces.xep0060_errors, "unsupported")

    feature = xso.Attr(
        "feature",
        validator=xso.RestrictToSet({
            "access-authorize",
            "access-open",
            "access-presence",
            "access-roster",
            "access-whitelist",
            "auto-create",
            "auto-subscribe",
            "collections",
            "config-node",
            "create-and-configure",
            "create-nodes",
            "delete-items",
            "delete-nodes",
            "filtered-notifications",
            "get-pending",
            "instant-nodes",
            "item-ids",
            "last-published",
            "leased-subscription",
            "manage-subscriptions",
            "member-affiliation",
            "meta-data",
            "modify-affiliations",
            "multi-collection",
            "multi-subscribe",
            "outcast-affiliation",
            "persistent-items",
            "presence-notifications",
            "presence-subscribe",
            "publish",
            "publish-options",
            "publish-only-affiliation",
            "publisher-affiliation",
            "purge-nodes",
            "retract-items",
            "retrieve-affiliations",
            "retrieve-default",
            "retrieve-items",
            "retrieve-subscriptions",
            "subscribe",
            "subscription-options",
            "subscription-notifications",
        })
    )


# foo
Feature.__docstring___ = \
    """
    .. attribute:: ACCESS_AUTHORIZE
       :annotation: = "http://jabber.org/protocol/pubsub#access-authorize"

       The default node access model is authorize.

    .. attribute:: ACCESS_OPEN
       :annotation: = "http://jabber.org/protocol/pubsub#access-open"

       The default node access model is open.

    .. attribute:: ACCESS_PRESENCE
       :annotation: = "http://jabber.org/protocol/pubsub#access-presence"

       The default node access model is presence.

    .. attribute:: ACCESS_ROSTER
       :annotation: = "http://jabber.org/protocol/pubsub#access-roster"

       The default node access model is roster.

    .. attribute:: ACCESS_WHITELIST
       :annotation: = "http://jabber.org/protocol/pubsub#access-whitelist"

       The default node access model is whitelist.

    .. attribute:: AUTO_CREATE
       :annotation: = "http://jabber.org/protocol/pubsub#auto-create"

       The service supports automatic creation of nodes on first publish.

    .. attribute:: AUTO_SUBSCRIBE
       :annotation: = "http://jabber.org/protocol/pubsub#auto-subscribe"

       The service supports automatic subscription to a nodes based on presence subscription.

    .. attribute:: COLLECTIONS
       :annotation: = "http://jabber.org/protocol/pubsub#collections"

       Collection nodes are supported.

    .. attribute:: CONFIG_NODE
       :annotation: = "http://jabber.org/protocol/pubsub#config-node"

       Configuration of node options is supported.

    .. attribute:: CREATE_AND_CONFIGURE
       :annotation: = "http://jabber.org/protocol/pubsub#create-and-configure"

       Simultaneous creation and configuration of nodes is supported.

    .. attribute:: CREATE_NODES
       :annotation: = "http://jabber.org/protocol/pubsub#create-nodes"

       Creation of nodes is supported.

    .. attribute:: DELETE_ITEMS
       :annotation: = "http://jabber.org/protocol/pubsub#delete-items"

       Deletion of items is supported.

    .. attribute:: DELETE_NODES
       :annotation: = "http://jabber.org/protocol/pubsub#delete-nodes"

       Deletion of nodes is supported.

    .. attribute:: FILTERED_NOTIFICATIONS
       :annotation: = "http://jabber.org/protocol/pubsub#filtered-notifications"

       The service supports filtering of notifications based on Entity Capabilities.

    .. attribute:: GET_PENDING
       :annotation: = "http://jabber.org/protocol/pubsub#get-pending"

       Retrieval of pending subscription approvals is supported.

    .. attribute:: INSTANT_NODES
       :annotation: = "http://jabber.org/protocol/pubsub#instant-nodes"

       Creation of instant nodes is supported.

    .. attribute:: ITEM_IDS
       :annotation: = "http://jabber.org/protocol/pubsub#item-ids"

       Publishers may specify item identifiers.

    .. attribute:: LAST_PUBLISHED
       :annotation: = "http://jabber.org/protocol/pubsub#last-published"

       The service supports sending of the last published item to new subscribers and to newly available resources.

    .. attribute:: LEASED_SUBSCRIPTION
       :annotation: = "http://jabber.org/protocol/pubsub#leased-subscription"

       Time-based subscriptions are supported.

    .. attribute:: MANAGE_SUBSCRIPTIONS
       :annotation: = "http://jabber.org/protocol/pubsub#manage-subscriptions"

       Node owners may manage subscriptions.

    .. attribute:: MEMBER_AFFILIATION
       :annotation: = "http://jabber.org/protocol/pubsub#member-affiliation"

       The member affiliation is supported.

    .. attribute:: META_DATA
       :annotation: = "http://jabber.org/protocol/pubsub#meta-data"

       Node meta-data is supported.

    .. attribute:: MODIFY_AFFILIATIONS
       :annotation: = "http://jabber.org/protocol/pubsub#modify-affiliations"

       Node owners may modify affiliations.

    .. attribute:: MULTI_COLLECTION
       :annotation: = "http://jabber.org/protocol/pubsub#multi-collection"

       A single leaf node can be associated with multiple collections.

    .. attribute:: MULTI_SUBSCRIBE
       :annotation: = "http://jabber.org/protocol/pubsub#multi-subscribe"

       A single entity may subscribe to a node multiple times.

    .. attribute:: OUTCAST_AFFILIATION
       :annotation: = "http://jabber.org/protocol/pubsub#outcast-affiliation"

       The outcast affiliation is supported.

    .. attribute:: PERSISTENT_ITEMS
       :annotation: = "http://jabber.org/protocol/pubsub#persistent-items"

       Persistent items are supported.

    .. attribute:: PRESENCE_NOTIFICATIONS
       :annotation: = "http://jabber.org/protocol/pubsub#presence-notifications"

       Presence-based delivery of event notifications is supported.

    .. attribute:: PRESENCE_SUBSCRIBE
       :annotation: = "http://jabber.org/protocol/pubsub#presence-subscribe"

       Implicit presence-based subscriptions are supported.

    .. attribute:: PUBLISH
       :annotation: = "http://jabber.org/protocol/pubsub#publish"

       Publishing items is supported.

    .. attribute:: PUBLISH_OPTIONS
       :annotation: = "http://jabber.org/protocol/pubsub#publish-options"

       Publication with publish options is supported.

    .. attribute:: PUBLISH_ONLY_AFFILIATION
       :annotation: = "http://jabber.org/protocol/pubsub#publish-only-affiliation"

       The publish-only affiliation is supported.

    .. attribute:: PUBLISHER_AFFILIATION
       :annotation: = "http://jabber.org/protocol/pubsub#publisher-affiliation"

       The publisher affiliation is supported.

    .. attribute:: PURGE_NODES
       :annotation: = "http://jabber.org/protocol/pubsub#purge-nodes"

       Purging of nodes is supported.

    .. attribute:: RETRACT_ITEMS
       :annotation: = "http://jabber.org/protocol/pubsub#retract-items"

       Item retraction is supported.

    .. attribute:: RETRIEVE_AFFILIATIONS
       :annotation: = "http://jabber.org/protocol/pubsub#retrieve-affiliations"

       Retrieval of current affiliations is supported.

    .. attribute:: RETRIEVE_DEFAULT
       :annotation: = "http://jabber.org/protocol/pubsub#retrieve-default"

       Retrieval of default node configuration is supported.

    .. attribute:: RETRIEVE_ITEMS
       :annotation: = "http://jabber.org/protocol/pubsub#retrieve-items"

       Item retrieval is supported.

    .. attribute:: RETRIEVE_SUBSCRIPTIONS
       :annotation: = "http://jabber.org/protocol/pubsub#retrieve-subscriptions"

       Retrieval of current subscriptions is supported.

    .. attribute:: SUBSCRIBE
       :annotation: = "http://jabber.org/protocol/pubsub#subscribe"

       Subscribing and unsubscribing are supported.

    .. attribute:: SUBSCRIPTION_OPTIONS
       :annotation: = "http://jabber.org/protocol/pubsub#subscription-options"

       Configuration of subscription options is supported.

    .. attribute:: SUBSCRIPTION_NOTIFICATIONS
       :annotation: = "http://jabber.org/protocol/pubsub#subscription-notifications"

       Notification of subscription state changes is supported.
    """



########################################################################
# File name: xso.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import aioxmpp.forms.xso as forms_xso
import aioxmpp.stanza as stanza
import aioxmpp.xso as xso

from aioxmpp.utils import namespaces

namespaces.xep0030_info = "http://jabber.org/protocol/disco#info"
namespaces.xep0030_items = "http://jabber.org/protocol/disco#items"


class Identity(xso.XSO):
    """
    An identity declaration. The keyword arguments to the constructor can be
    used to initialize attributes of the :class:`Identity` instance.

    .. attribute:: category

       The category of the identity. The value is not validated against the
       values in the `registry
       <https://xmpp.org/registrar/disco-categories.html>`_.

    .. attribute:: type_

       The type of the identity. The value is not validated against the values
       in the `registry
       <https://xmpp.org/registrar/disco-categories.html>`_.

    .. attribute:: name

       The optional human-readable name of the identity. See also the
       :attr:`lang` attribute.

    .. attribute:: lang

       The language of the :attr:`name`. This may be not :data:`None` even if
       :attr:`name` is not set due to ``xml:lang`` propagation.

    """
    TAG = (namespaces.xep0030_info, "identity")

    category = xso.Attr(tag="category")
    type_ = xso.Attr(tag="type")
    name = xso.Attr(tag="name", default=None)
    lang = xso.LangAttr()

    def __init__(self, *,
                 category="client",
                 type_="bot",
                 name=None,
                 lang=None):
        super().__init__()
        self.category = category
        self.type_ = type_
        if name is not None:
            self.name = name
        if lang is not None:
            self.lang = lang

    def __eq__(self, other):
        try:
            return (self.category == other.category and
                    self.type_ == other.type_ and
                    self.name == other.name and
                    self.lang == other.lang)
        except AttributeError:
            return NotImplemented

    def __repr__(self):
        return "{}.{}(category={!r}, type_={!r}, name={!r}, lang={!r})".format(
            self.__class__.__module__,
            self.__class__.__qualname__,
            self.category,
            self.type_,
            self.name,
            self.lang)


class Feature(xso.XSO):
    """
    A feature declaration. The keyword argument to the constructor can be used
    to initialize the attribute of the :class:`Feature` instance.

    .. attribute:: var

       The namespace which identifies the feature.

    """

    TAG = (namespaces.xep0030_info, "feature")

    var = xso.Attr(tag="var")

    def __init__(self, var):
        super().__init__()
        self.var = var


class FeatureSet(xso.AbstractElementType):
    def get_xso_types(self):
        return [Feature]

    def unpack(self, item):
        return item.var

    def pack(self, var):
        return Feature(var)


@stanza.IQ.as_payload_class
class InfoQuery(xso.CapturingXSO):
    """
    A query for features and identities of an entity. The keyword arguments to
    the constructor can be used to initialize the attributes. Note that
    `identities` and `features` must be iterables of :class:`Identity` and
    :class:`Feature`, respectively; these iterables are evaluated and the items
    are stored in the respective attributes.

    .. attribute:: node

       The node at which the query is directed.

    .. attribute:: identities

       The identities of the entity, as :class:`Identity` instances. Each
       entity has at least one identity.

    .. attribute:: features

       The features of the entity, as a set of strings. Each string represents
       a :class:`Feature` instance with the corresponding :attr:`~.Feature.var`
       attribute.

    .. attribute:: captured_events

       If the object was created by parsing an XML stream, this attribute holds
       a list of events which were used when parsing it.

       Otherwise, this is :data:`None`.

       .. versionadded:: 0.5

    .. automethod:: to_dict

    """
    __slots__ = ("captured_events",)

    TAG = (namespaces.xep0030_info, "query")

    node = xso.Attr(tag="node", default=None)

    identities = xso.ChildList([Identity])

    features = xso.ChildValueList(
        FeatureSet(),
        container_type=set
    )

    exts = xso.ChildList([forms_xso.Data])

    def __init__(self, *, identities=(), features=(), node=None):
        super().__init__()
        self.captured_events = None
        self.identities.extend(identities)
        self.features.update(features)
        if node is not None:
            self.node = node

    def to_dict(self):
        """
        Convert the query result to a normalized JSON-like
        representation.

        The format is a subset of the format used by the `capsdb`__. Obviously,
        the node name and hash type are not included; otherwise, the format is
        identical.

        __ https://github.com/xnyhps/capsdb
        """
        identities = []
        for identity in self.identities:
            identity_dict = {
                "category": identity.category,
                "type": identity.type_,
            }
            if identity.lang is not None:
                identity_dict["lang"] = identity.lang.match_str
            if identity.name is not None:
                identity_dict["name"] = identity.name
            identities.append(identity_dict)

        features = sorted(self.features)

        forms = []
        for form in self.exts:
            forms.append({
                field.var: list(field.values)
                for field in form.fields
                if field.var is not None
            })

        result = {
            "identities": identities,
            "features": features,
            "forms": forms
        }

        return result

    def _set_captured_events(self, events):
        self.captured_events = events


class Item(xso.XSO):
    """
    An item declaration. The keyword arguments to the constructor can be used
    to initialize the attributes of the :class:`Item` instance.

    .. attribute:: jid

       :class:`~aioxmpp.JID` of the entity represented by the item.

    .. attribute:: node

       Node of the item

    .. attribute:: name

       Name of the item

    """

    TAG = (namespaces.xep0030_items, "item")
    UNKNOWN_CHILD_POLICY = xso.UnknownChildPolicy.DROP

    jid = xso.Attr(
        tag="jid",
        type_=xso.JID(),
        # FIXME: validator for full jid
    )

    name = xso.Attr(
        tag="name",
        default=None,
    )

    node = xso.Attr(
        tag="node",
        default=None,
    )

    def __init__(self, jid, name=None, node=None):
        super().__init__()
        self.jid = jid
        self.name = name
        self.node = node


@stanza.IQ.as_payload_class
class ItemsQuery(xso.XSO):
    """
    A query for items at a specific entity. The keyword arguments to the
    constructor can be used to initialize the attributes of the
    :class:`ItemsQuery`. Note that `items` must be an iterable of :class:`Item`
    instances. The iterable will be evaluated and the items will be stored in
    the :attr:`items` attribute.

    .. attribute:: node

       Node at which the query is directed

    .. attribute:: items

       The items at the addressed entity.

    """
    TAG = (namespaces.xep0030_items, "query")

    node = xso.Attr(tag="node", default=None)

    items = xso.ChildList([Item])

    def __init__(self, *, node=None, items=()):
        super().__init__()
        self.items.extend(items)
        if node is not None:
            self.node = node



########################################################################
# File name: aioxmppspecific.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio

from sphinx import addnodes
from sphinx.domains.python import PyModulelevel, PyClassmember

from docutils import nodes, utils
from docutils.parsers.rst import roles

from sphinx.locale import _

from sphinx.environment import default_settings

from sphinx.ext.autodoc import (
    MethodDocumenter,
    FunctionDocumenter,
    ModuleLevelDocumenter
)

from sphinx.util.nodes import (
    split_explicit_title
)


class PyCoroutineMixin(object):
    def handle_signature(self, sig, signode):
        ret = super(PyCoroutineMixin, self).handle_signature(sig, signode)
        signode.insert(0, addnodes.desc_annotation('coroutine ', 'coroutine '))
        return ret


class PyCoroutineFunction(PyCoroutineMixin, PyModulelevel):
    def run(self):
        self.name = 'py:function'
        return PyModulelevel.run(self)


class PyCoroutineMethod(PyCoroutineMixin, PyClassmember):
    def run(self):
        self.name = 'py:method'
        return PyClassmember.run(self)


class CoroutineAwareFunctionDocumenter(FunctionDocumenter):
    objtype = 'function'
    priority = 3

    def import_object(self):
        ret = ModuleLevelDocumenter.import_object(self)
        if not ret:
            return ret

        if asyncio.iscoroutinefunction(self.object):
            self.directivetype = "coroutinefunction"

        return ret


class DecoratorDocumenter(FunctionDocumenter):
    objtype = 'decorator'
    priority = 3


class CoroutineAwareMethodDocumenter(MethodDocumenter):
    objtype = 'method'
    priority = 4

    def import_object(self):
        ret = super().import_object()
        if not ret:
            return ret

        if     (self.directivetype == "method" and
                asyncio.iscoroutinefunction(self.object)):
            self.directivetype = "coroutinemethod"

        return ret


class SignalAwareMethodDocumenter(MethodDocumenter):
    objtype = 'signal'
    priority = 4

    def import_object(self):
        ret = super().import_object()
        if not ret:
            return ret

        self.directivetype = "signal"

        return ret


class PySignal(PyClassmember):
    def handle_signature(self, sig, signode):
        ret = super(PySignal, self).handle_signature(sig, signode)
        signode.insert(0, addnodes.desc_annotation('signal ', 'signal '))
        return ret

    def run(self):
        self.name = 'py:method'
        return PyClassmember.run(self)


class PySyncSignal(PyClassmember):
    def handle_signature(self, sig, signode):
        ret = super(PySyncSignal, self).handle_signature(sig, signode)
        signode.insert(0, addnodes.desc_annotation('coroutine signal ', 'coroutine signal '))
        return ret

    def run(self):
        self.name = 'py:method'
        return PyClassmember.run(self)


def xep_role(typ, rawtext, text, lineno, inliner,
             options={}, content=[]):
    """Role for PEP/RFC references that generate an index entry."""
    env = inliner.document.settings.env
    if not typ:
        typ = env.config.default_role
    else:
        typ = typ.lower()
    has_explicit_title, title, target = split_explicit_title(text)
    title = utils.unescape(title)
    target = utils.unescape(target)
    targetid = 'index-%s' % env.new_serialno('index')
    anchor = ''
    anchorindex = target.find('#')
    if anchorindex > 0:
        target, anchor = target[:anchorindex], target[anchorindex:]
    try:
        xepnum = int(target)
    except ValueError:
        msg = inliner.reporter.error('invalid XEP number %s' % target,
                                     line=lineno)
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]
    target = "{:04d}".format(xepnum)
    if not has_explicit_title:
        title = "XEP-" + target
    indexnode = addnodes.index()
    targetnode = nodes.target('', '', ids=[targetid])
    inliner.document.note_explicit_target(targetnode)
    indexnode['entries'] = [
        ('single', _('XMPP Extension Protocols (XEPs); XEP %s') % target,
         targetid, '')]
    ref = inliner.document.settings.xep_base_url + 'xep-%04d.html' % xepnum
    rn = nodes.reference(title, title, internal=False, refuri=ref+anchor,
                         classes=[typ])
    return [indexnode, targetnode, rn], []


roles.register_local_role("xep", xep_role)


default_settings["xep_base_url"] = "https://xmpp.org/extensions/"


def setup(app):
    app.add_directive_to_domain('py', 'coroutinefunction', PyCoroutineFunction)
    app.add_directive_to_domain('py', 'coroutinemethod', PyCoroutineMethod)
    app.add_directive_to_domain('py', 'signal', PySignal)
    app.add_directive_to_domain('py', 'syncsignal', PySyncSignal)
    app.add_autodocumenter(CoroutineAwareFunctionDocumenter)
    app.add_autodocumenter(CoroutineAwareMethodDocumenter)
    app.add_autodocumenter(SignalAwareMethodDocumenter)
    app.add_autodocumenter(DecoratorDocumenter)
    return {'version': '1.0', 'parallel_read_safe': True}



########################################################################
# File name: set_avatar.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio

import aioxmpp
import aioxmpp.avatar

from framework import Example, exec_example


class Avatar(Example):
    def prepare_argparse(self):
        super().prepare_argparse()

        group = self.argparse.add_mutually_exclusive_group(required=True)

        group.add_argument(
            "--set-avatar", nargs=1, metavar="AVATAR_FILE",
            help="set the avatar to content of the supplied PNG file."
        )

        group.add_argument(
            "--wipe-avatar",
            action="store_true",
            default=False,
            help="set the avatar to no avatar."
        )

    def configure(self):
        super().configure()
        if self.args.set_avatar:
            self.avatar_file, = self.args.set_avatar
        else:
            self.avatar_file = None
        self.wipe_avatar = self.args.wipe_avatar

    def make_simple_client(self):
        client = super().make_simple_client()
        self.avatar = client.summon(aioxmpp.avatar.AvatarService)
        return client

    @asyncio.coroutine
    def run_simple_example(self):
        if self.avatar_file is not None:
            with open(self.avatar_file, "rb") as f:
                image_data = f.read()

            avatar_set = aioxmpp.avatar.AvatarSet()
            avatar_set.add_avatar_image("image/png", image_bytes=image_data)

            yield from self.avatar.publish_avatar_set(avatar_set)

        elif self.wipe_avatar:
            yield from self.avatar.disable_avatar()

    @asyncio.coroutine
    def run_example(self):
        yield from super().run_example()


if __name__ == "__main__":
    exec_example(Avatar())



########################################################################
# File name: test_xso.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import contextlib
import unittest
import unittest.mock

import aioxmpp.forms as forms
import aioxmpp.pubsub.xso as pubsub_xso
import aioxmpp.stanza as stanza
import aioxmpp.structs as structs
import aioxmpp.xso as xso

from aioxmpp.utils import namespaces


TEST_JID = structs.JID.fromstr("foo@bar.example/baz")


@pubsub_xso.as_payload_class
class SomePayload(xso.XSO):
    TAG = "aioxmpp.tests.pubsub.test_xso", "foo"


class TestNamespaces(unittest.TestCase):
    def test_features(self):
        self.assertIs(
            namespaces.xep0060_features,
            pubsub_xso.Feature
        )

    def test_core(self):
        self.assertEqual(
            namespaces.xep0060,
            "http://jabber.org/protocol/pubsub"
        )

    def test_errors(self):
        self.assertEqual(
            namespaces.xep0060_errors,
            "http://jabber.org/protocol/pubsub#errors"
        )

    def test_event(self):
        self.assertEqual(
            namespaces.xep0060_event,
            "http://jabber.org/protocol/pubsub#event"
        )

    def test_owner(self):
        self.assertEqual(
            namespaces.xep0060_owner,
            "http://jabber.org/protocol/pubsub#owner"
        )


class TestAffiliation(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.Affiliation,
            xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.Affiliation.TAG,
            (namespaces.xep0060, "affiliation")
        )

    def test_node(self):
        self.assertIsInstance(
            pubsub_xso.Affiliation.node,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.Affiliation.node.tag,
            (None, "node")
        )
        self.assertIsInstance(
            pubsub_xso.Affiliation.node.type_,
            xso.String
        )
        self.assertIs(
            pubsub_xso.Affiliation.node.default,
            None
        )

    def test_affiliation(self):
        self.assertIsInstance(
            pubsub_xso.Affiliation.affiliation,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.Affiliation.affiliation.tag,
            (None, "affiliation")
        )
        self.assertIsInstance(
            pubsub_xso.Affiliation.affiliation.type_,
            xso.String
        )
        self.assertIsInstance(
            pubsub_xso.Affiliation.affiliation.validator,
            xso.RestrictToSet
        )
        self.assertSetEqual(
            pubsub_xso.Affiliation.affiliation.validator.values,
            {
                "member",
                "none",
                "outcast",
                "owner",
                "publisher",
                "publish-only",
            }
        )
        self.assertIs(
            pubsub_xso.Affiliation.affiliation.default,
            xso.NO_DEFAULT,
        )

    def test_init(self):
        with self.assertRaises(TypeError):
            a = pubsub_xso.Affiliation()

        a = pubsub_xso.Affiliation("owner")
        self.assertEqual(a.node, None)
        self.assertEqual(a.affiliation, "owner")

        a = pubsub_xso.Affiliation("outcast", node="foo")
        self.assertEqual(a.node, "foo")
        self.assertEqual(a.affiliation, "outcast")


class TestAffiliations(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.Affiliations,
            xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.Affiliations.TAG,
            (namespaces.xep0060, "affiliations")
        )

    def test_node(self):
        self.assertIsInstance(
            pubsub_xso.Affiliations.node,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.Affiliations.node.tag,
            (None, "node")
        )
        self.assertIsInstance(
            pubsub_xso.Affiliations.node.type_,
            xso.String
        )
        self.assertIs(
            pubsub_xso.Affiliations.node.default,
            None
        )

    def test_affiliations(self):
        self.assertIsInstance(
            pubsub_xso.Affiliations.affiliations,
            xso.ChildList
        )
        self.assertSetEqual(
            pubsub_xso.Affiliations.affiliations._classes,
            {
                pubsub_xso.Affiliation
            }
        )

    def test_init(self):
        as_ = pubsub_xso.Affiliations()
        self.assertIsNone(as_.node)
        self.assertSequenceEqual(as_.affiliations, [])

        a1 = pubsub_xso.Affiliation("owner")
        a2 = pubsub_xso.Affiliation("member")
        as_ = pubsub_xso.Affiliations([a1, a2], node="foo")
        self.assertEqual(as_.node, "foo")
        self.assertSequenceEqual(
            as_.affiliations,
            [
                a1, a2
            ]
        )


class TestConfigure(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.Configure,
            xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.Configure.TAG,
            (namespaces.xep0060, "configure")
        )

    def test_data(self):
        self.assertIsInstance(
            pubsub_xso.Configure.data,
            xso.Child,
        )
        self.assertSetEqual(
            pubsub_xso.Configure.data._classes,
            {
                forms.Data,
            }
        )


class TestCreate(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.Create,
            xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.Create.TAG,
            (namespaces.xep0060, "create")
        )

    def test_node(self):
        self.assertIsInstance(
            pubsub_xso.Create.node,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.Create.node.tag,
            (None, "node")
        )
        self.assertIsInstance(
            pubsub_xso.Create.node.type_,
            xso.String
        )
        self.assertIs(
            pubsub_xso.Create.node.default,
            None
        )


class TestDefault(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.Default,
            xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.Default.TAG,
            (namespaces.xep0060, "default")
        )

    def test_node(self):
        self.assertIsInstance(
            pubsub_xso.Default.node,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.Default.node.tag,
            (None, "node")
        )
        self.assertIsInstance(
            pubsub_xso.Default.node.type_,
            xso.String
        )
        self.assertIs(
            pubsub_xso.Default.node.default,
            None
        )

    def test_type_(self):
        self.assertIsInstance(
            pubsub_xso.Default.type_,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.Default.type_.tag,
            (None, "type")
        )
        self.assertIsInstance(
            pubsub_xso.Default.type_.type_,
            xso.String
        )
        self.assertIs(
            pubsub_xso.Default.type_.default,
            "leaf",
        )
        self.assertIsInstance(
            pubsub_xso.Default.type_.validator,
            xso.RestrictToSet
        )
        self.assertSetEqual(
            pubsub_xso.Default.type_.validator.values,
            {
                "leaf",
                "collection",
            }
        )

    def test_data(self):
        self.assertIsInstance(
            pubsub_xso.Default.data,
            xso.Child,
        )
        self.assertSetEqual(
            pubsub_xso.Default.data._classes,
            {
                forms.Data,
            }
        )

    def test_init(self):
        d = pubsub_xso.Default()
        self.assertIsNone(d.node)
        self.assertEqual(d.type_, "leaf")
        self.assertIsNone(d.data)

        data = unittest.mock.Mock()
        d = pubsub_xso.Default(node="foo", data=data)
        self.assertEqual(d.node, "foo")
        self.assertIs(d.data, data)


class TestItem(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.Item,
            xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.Item.TAG,
            (namespaces.xep0060, "item")
        )

    def test_id_(self):
        self.assertIsInstance(
            pubsub_xso.Item.id_,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.Item.id_.tag,
            (None, "id")
        )
        self.assertIs(
            pubsub_xso.Item.id_.default,
            None
        )

    def test_registered_payload(self):
        self.assertIsInstance(
            pubsub_xso.Item.registered_payload,
            xso.Child
        )

    def test_unknown_payload(self):
        self.assertIsInstance(
            pubsub_xso.Item.unregistered_payload,
            xso.Collector
        )

    def test_init(self):
        i = pubsub_xso.Item()
        self.assertIsNone(i.id_)
        self.assertFalse(i.registered_payload)
        self.assertSequenceEqual(i.unregistered_payload, [])

        i = pubsub_xso.Item("foo")
        self.assertEqual(i.id_, "foo")


class TestItems(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.Items,
            xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.Items.TAG,
            (namespaces.xep0060, "items")
        )

    def test_max_items(self):
        self.assertIsInstance(
            pubsub_xso.Items.max_items,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.Items.max_items.tag,
            (None, "max_items"),
        )
        self.assertIsInstance(
            pubsub_xso.Items.max_items.type_,
            xso.Integer
        )
        self.assertIsInstance(
            pubsub_xso.Items.max_items.validator,
            xso.NumericRange
        )
        self.assertEqual(
            pubsub_xso.Items.max_items.validator.min_,
            1
        )
        self.assertIs(
            pubsub_xso.Items.max_items.validator.max_,
            None
        )
        self.assertIs(
            pubsub_xso.Items.max_items.default,
            None
        )

    def test_node(self):
        self.assertIsInstance(
            pubsub_xso.Items.node,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.Items.node.tag,
            (None, "node")
        )
        self.assertIs(
            pubsub_xso.Items.node.default,
            xso.NO_DEFAULT,
        )

    def test_subid(self):
        self.assertIsInstance(
            pubsub_xso.Items.subid,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.Items.subid.tag,
            (None, "subid")
        )
        self.assertIsInstance(
            pubsub_xso.Items.subid.type_,
            xso.String
        )
        self.assertIs(
            pubsub_xso.Items.subid.default,
            None
        )

    def test_items(self):
        self.assertIsInstance(
            pubsub_xso.Items.items,
            xso.ChildList
        )
        self.assertSetEqual(
            pubsub_xso.Items.items._classes,
            {
                pubsub_xso.Item
            }
        )

    def test_init(self):
        with self.assertRaises(TypeError):
            i = pubsub_xso.Items()

        i = pubsub_xso.Items("foo")
        self.assertEqual(i.node, "foo")
        self.assertIsNone(i.max_items)
        self.assertIsNone(i.subid)

        i = pubsub_xso.Items("foo", max_items=10, subid="bar")
        self.assertEqual(i.node, "foo")
        self.assertEqual(i.max_items, 10)
        self.assertEqual(i.subid, "bar")


class TestOptions(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.Options,
            xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.Options.TAG,
            (namespaces.xep0060, "options")
        )

    def test_jid(self):
        self.assertIsInstance(
            pubsub_xso.Options.jid,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.Options.jid.tag,
            (None, "jid")
        )
        self.assertIsInstance(
            pubsub_xso.Options.jid.type_,
            xso.JID
        )
        self.assertIs(
            pubsub_xso.Options.jid.default,
            xso.NO_DEFAULT
        )

    def test_node(self):
        self.assertIsInstance(
            pubsub_xso.Options.node,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.Options.node.tag,
            (None, "node")
        )
        self.assertIsInstance(
            pubsub_xso.Options.node.type_,
            xso.String
        )
        self.assertIs(
            pubsub_xso.Options.node.default,
            None
        )

    def test_subid(self):
        self.assertIsInstance(
            pubsub_xso.Options.subid,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.Options.subid.tag,
            (None, "subid")
        )
        self.assertIsInstance(
            pubsub_xso.Options.subid.type_,
            xso.String
        )
        self.assertIs(
            pubsub_xso.Options.subid.default,
            None
        )

    def test_data(self):
        self.assertIsInstance(
            pubsub_xso.Options.data,
            xso.Child,
        )
        self.assertSetEqual(
            pubsub_xso.Options.data._classes,
            {
                forms.Data,
            }
        )

    def test_init(self):
        with self.assertRaises(TypeError):
            o = pubsub_xso.Options()

        o = pubsub_xso.Options(TEST_JID)
        self.assertEqual(o.jid, TEST_JID)
        self.assertIsNone(o.node)
        self.assertIsNone(o.subid)
        self.assertIsNone(o.data)

        o = pubsub_xso.Options(TEST_JID, node="foo", subid="bar")
        self.assertEqual(o.jid, TEST_JID)
        self.assertEqual(o.node, "foo")
        self.assertEqual(o.subid, "bar")
        self.assertIsNone(o.data)


class TestPublish(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.Publish,
            xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.Publish.TAG,
            (namespaces.xep0060, "publish")
        )

    def test_node(self):
        self.assertIsInstance(
            pubsub_xso.Publish.node,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.Publish.node.tag,
            (None, "node")
        )
        self.assertIs(
            pubsub_xso.Publish.node.default,
            xso.NO_DEFAULT,
        )

    def test_item(self):
        self.assertIsInstance(
            pubsub_xso.Publish.item,
            xso.Child,
        )
        self.assertSetEqual(
            pubsub_xso.Publish.item._classes,
            {
                pubsub_xso.Item,
            }
        )


class TestRetract(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.Retract,
            xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.Retract.TAG,
            (namespaces.xep0060, "retract")
        )

    def test_node(self):
        self.assertIsInstance(
            pubsub_xso.Retract.node,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.Retract.node.tag,
            (None, "node")
        )
        self.assertIs(
            pubsub_xso.Retract.node.default,
            xso.NO_DEFAULT,
        )

    def test_item(self):
        self.assertIsInstance(
            pubsub_xso.Retract.item,
            xso.Child,
        )
        self.assertSetEqual(
            pubsub_xso.Retract.item._classes,
            {
                pubsub_xso.Item,
            }
        )

    def test_notify(self):
        self.assertIsInstance(
            pubsub_xso.Retract.notify,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.Retract.notify.tag,
            (None, "notify")
        )
        self.assertIsInstance(
            pubsub_xso.Retract.notify.type_,
            xso.Bool
        )
        self.assertIs(
            pubsub_xso.Retract.notify.default,
            False
        )


class TestSubscribe(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.Subscribe,
            xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.Subscribe.TAG,
            (namespaces.xep0060, "subscribe")
        )

    def test_jid(self):
        self.assertIsInstance(
            pubsub_xso.Subscribe.jid,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.Subscribe.jid.tag,
            (None, "jid")
        )
        self.assertIsInstance(
            pubsub_xso.Subscribe.jid.type_,
            xso.JID
        )
        self.assertIs(
            pubsub_xso.Subscribe.jid.default,
            xso.NO_DEFAULT
        )

    def test_node(self):
        self.assertIsInstance(
            pubsub_xso.Subscribe.node,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.Subscribe.node.tag,
            (None, "node")
        )
        self.assertIsInstance(
            pubsub_xso.Subscribe.node.type_,
            xso.String
        )
        self.assertIs(
            pubsub_xso.Subscribe.node.default,
            None
        )

    def test_init(self):
        with self.assertRaises(TypeError):
            s = pubsub_xso.Subscribe()

        s = pubsub_xso.Subscribe(TEST_JID)
        self.assertEqual(s.jid, TEST_JID)
        self.assertEqual(s.node, None)

        s = pubsub_xso.Subscribe(
            TEST_JID.replace(localpart="fnord"),
            node="foo")
        self.assertEqual(s.jid, TEST_JID.replace(localpart="fnord"))
        self.assertEqual(s.node, "foo")


class TestSubscribeOptions(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.SubscribeOptions,
            xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.SubscribeOptions.TAG,
            (namespaces.xep0060, "subscribe-options")
        )

    def test_required(self):
        self.assertIsInstance(
            pubsub_xso.SubscribeOptions.required,
            xso.ChildTag
        )
        self.assertSetEqual(
            set(pubsub_xso.SubscribeOptions.required.get_tag_map()),
            {
                (namespaces.xep0060, "required"),
            }
        )
        self.assertIs(
            pubsub_xso.SubscribeOptions.required.allow_none,
            True
        )


class TestSubscription(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.Subscription,
            xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.Subscription.TAG,
            (namespaces.xep0060, "subscription")
        )

    def test_jid(self):
        self.assertIsInstance(
            pubsub_xso.Subscription.jid,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.Subscription.jid.tag,
            (None, "jid")
        )
        self.assertIsInstance(
            pubsub_xso.Subscription.jid.type_,
            xso.JID
        )
        self.assertIs(
            pubsub_xso.Subscription.jid.default,
            xso.NO_DEFAULT
        )

    def test_node(self):
        self.assertIsInstance(
            pubsub_xso.Subscription.node,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.Subscription.node.tag,
            (None, "node")
        )
        self.assertIsInstance(
            pubsub_xso.Subscription.node.type_,
            xso.String
        )
        self.assertIs(
            pubsub_xso.Subscription.node.default,
            None
        )

    def test_subid(self):
        self.assertIsInstance(
            pubsub_xso.Subscription.subid,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.Subscription.subid.tag,
            (None, "subid")
        )
        self.assertIsInstance(
            pubsub_xso.Subscription.subid.type_,
            xso.String
        )
        self.assertIs(
            pubsub_xso.Subscription.subid.default,
            None
        )

    def test_subscription(self):
        self.assertIsInstance(
            pubsub_xso.Subscription.subscription,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.Subscription.subscription.tag,
            (None, "subscription")
        )
        self.assertIsInstance(
            pubsub_xso.Subscription.subscription.type_,
            xso.String
        )
        self.assertIsInstance(
            pubsub_xso.Subscription.subscription.validator,
            xso.RestrictToSet
        )
        self.assertSetEqual(
            pubsub_xso.Subscription.subscription.validator.values,
            {
                "none",
                "pending",
                "subscribed",
                "unconfigured"
            }
        )
        self.assertIs(
            pubsub_xso.Subscription.subscription.default,
            None
        )

    def test_subscribe_options(self):
        self.assertIsInstance(
            pubsub_xso.Subscription.subscribe_options,
            xso.Child
        )
        self.assertSetEqual(
            pubsub_xso.Subscription.subscribe_options._classes,
            {
                pubsub_xso.SubscribeOptions
            }
        )

    def test_init(self):
        with self.assertRaises(TypeError):
            s = pubsub_xso.Subscription()

        s = pubsub_xso.Subscription(TEST_JID)
        self.assertEqual(s.jid, TEST_JID)
        self.assertIsNone(s.node)
        self.assertIsNone(s.subid)
        self.assertIsNone(s.subscription)

        s = pubsub_xso.Subscription(
            TEST_JID.replace(localpart="fnord"),
            node="foo",
            subid="bar",
            subscription="subscribed")
        self.assertEqual(s.jid, TEST_JID.replace(localpart="fnord"))
        self.assertEqual(s.node, "foo")
        self.assertEqual(s.subid, "bar")
        self.assertEqual(s.subscription, "subscribed")


class TestSubscriptions(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.Subscriptions,
            xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.Subscriptions.TAG,
            (namespaces.xep0060, "subscriptions"),
        )

    def test_node(self):
        self.assertIsInstance(
            pubsub_xso.Subscriptions.node,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.Subscriptions.node.tag,
            (None, "node")
        )
        self.assertIsInstance(
            pubsub_xso.Subscriptions.node.type_,
            xso.String
        )
        self.assertIs(
            pubsub_xso.Subscriptions.node.default,
            None
        )

    def test_subscriptions(self):
        self.assertIsInstance(
            pubsub_xso.Subscriptions.subscriptions,
            xso.ChildList
        )
        self.assertSetEqual(
            pubsub_xso.Subscriptions.subscriptions._classes,
            {
                pubsub_xso.Subscription
            }
        )

    def test_init_default(self):
        subs = pubsub_xso.Subscriptions()
        self.assertIsNone(subs.node)
        self.assertSequenceEqual(subs.subscriptions, [])

    def test_init(self):
        subs = pubsub_xso.Subscriptions(
            node="foobar",
            subscriptions=[
                unittest.mock.sentinel.s1,
                unittest.mock.sentinel.s2,
            ]
        )
        self.assertEqual(subs.node, "foobar")

        self.assertSequenceEqual(
            subs.subscriptions,
            [
                unittest.mock.sentinel.s1,
                unittest.mock.sentinel.s2,
            ]
        )


class TestUnsubscribe(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.Unsubscribe,
            xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.Unsubscribe.TAG,
            (namespaces.xep0060, "unsubscribe")
        )

    def test_jid(self):
        self.assertIsInstance(
            pubsub_xso.Unsubscribe.jid,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.Unsubscribe.jid.tag,
            (None, "jid")
        )
        self.assertIsInstance(
            pubsub_xso.Unsubscribe.jid.type_,
            xso.JID
        )
        self.assertIs(
            pubsub_xso.Unsubscribe.jid.default,
            xso.NO_DEFAULT
        )

    def test_node(self):
        self.assertIsInstance(
            pubsub_xso.Unsubscribe.node,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.Unsubscribe.node.tag,
            (None, "node")
        )
        self.assertIsInstance(
            pubsub_xso.Unsubscribe.node.type_,
            xso.String
        )
        self.assertIs(
            pubsub_xso.Unsubscribe.node.default,
            None
        )

    def test_subid(self):
        self.assertIsInstance(
            pubsub_xso.Unsubscribe.subid,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.Unsubscribe.subid.tag,
            (None, "subid")
        )
        self.assertIsInstance(
            pubsub_xso.Unsubscribe.subid.type_,
            xso.String
        )
        self.assertIs(
            pubsub_xso.Unsubscribe.subid.default,
            None
        )

    def test_init(self):
        with self.assertRaises(TypeError):
            u = pubsub_xso.Unsubscribe()

        u = pubsub_xso.Unsubscribe(TEST_JID)
        self.assertEqual(u.jid, TEST_JID)
        self.assertIsNone(u.node)
        self.assertIsNone(u.subid)

        u = pubsub_xso.Unsubscribe(
            TEST_JID.replace(localpart="fnord"),
            node="foo",
            subid="bar",
        )
        self.assertEqual(u.jid, TEST_JID.replace(localpart="fnord"))
        self.assertEqual(u.node, "foo")
        self.assertEqual(u.subid, "bar")


class TestRequest(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.Request,
            xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.Request.TAG,
            (namespaces.xep0060, "pubsub")
        )

    def test_payload(self):
        self.assertIsInstance(
            pubsub_xso.Request.payload,
            xso.Child
        )
        self.assertSetEqual(
            pubsub_xso.Request.payload._classes,
            {
                pubsub_xso.Affiliations,
                pubsub_xso.Create,
                pubsub_xso.Default,
                pubsub_xso.Items,
                pubsub_xso.Publish,
                pubsub_xso.Retract,
                pubsub_xso.Subscribe,
                pubsub_xso.Subscription,
                pubsub_xso.Subscriptions,
                pubsub_xso.Unsubscribe,
            }
        )

    def test_options(self):
        self.assertIsInstance(
            pubsub_xso.Request.options,
            xso.Child
        )
        self.assertSetEqual(
            pubsub_xso.Request.options._classes,
            {
                pubsub_xso.Options
            }
        )

    def test_configure(self):
        self.assertIsInstance(
            pubsub_xso.Request.configure,
            xso.Child
        )
        self.assertSetEqual(
            pubsub_xso.Request.configure._classes,
            {
                pubsub_xso.Configure
            }
        )

    def test_is_registered_iq_payload(self):
        self.assertIn(
            pubsub_xso.Request,
            stanza.IQ.payload._classes
        )

    def test_Message_attribute(self):
        self.assertIsInstance(
            stanza.Message.xep0060_request,
            xso.Child,
        )
        self.assertSetEqual(
            stanza.Message.xep0060_request._classes,
            {
                pubsub_xso.Request,
            }
        )

    def test_init(self):
        r = pubsub_xso.Request()
        self.assertIsNone(r.payload)

        m = unittest.mock.Mock()
        r = pubsub_xso.Request(m)
        self.assertIs(r.payload, m)


class TestEventAssociate(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.EventAssociate,
            xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.EventAssociate.TAG,
            (namespaces.xep0060_event, "associate"),
        )

    def test_node(self):
        self.assertIsInstance(
            pubsub_xso.EventAssociate.node,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.EventAssociate.node.tag,
            (None, "node")
        )
        self.assertIsInstance(
            pubsub_xso.EventAssociate.node.type_,
            xso.String
        )
        self.assertIs(
            pubsub_xso.EventAssociate.node.default,
            xso.NO_DEFAULT,
        )


class TestEventDisassociate(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.EventDisassociate,
            xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.EventDisassociate.TAG,
            (namespaces.xep0060_event, "disassociate"),
        )

    def test_node(self):
        self.assertIsInstance(
            pubsub_xso.EventDisassociate.node,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.EventDisassociate.node.tag,
            (None, "node")
        )
        self.assertIsInstance(
            pubsub_xso.EventDisassociate.node.type_,
            xso.String
        )
        self.assertIs(
            pubsub_xso.EventDisassociate.node.default,
            xso.NO_DEFAULT,
        )


class TestEventCollection(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.EventCollection,
            xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.EventCollection.TAG,
            (namespaces.xep0060_event, "collection"),
        )

    def test_assoc(self):
        self.assertIsInstance(
            pubsub_xso.EventCollection.assoc,
            xso.Child,
        )
        self.assertSetEqual(
            set(pubsub_xso.EventCollection.assoc._classes),
            {
                pubsub_xso.EventAssociate,
                pubsub_xso.EventDisassociate,
            }
        )


class TestEventRedirect(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.EventRedirect,
            xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.EventRedirect.TAG,
            (namespaces.xep0060_event, "redirect")
        )

    def test_uri(self):
        self.assertIsInstance(
            pubsub_xso.EventRedirect.uri,
            xso.Attr,
        )
        self.assertEqual(
            pubsub_xso.EventRedirect.uri.tag,
            (None, "uri"),
        )

    def test_init_default(self):
        with self.assertRaises(TypeError):
            pubsub_xso.EventRedirect()

    def test_init(self):
        r = pubsub_xso.EventRedirect("foobar")
        self.assertEqual(r.uri, "foobar")


class TestEventDelete(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.EventDelete,
            xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.EventDelete.TAG,
            (namespaces.xep0060_event, "delete"),
        )

    def test__redirect(self):
        self.assertIsInstance(
            pubsub_xso.EventDelete._redirect,
            xso.Child,
        )
        self.assertSetEqual(
            pubsub_xso.EventDelete._redirect._classes,
            {
                pubsub_xso.EventRedirect,
            }
        )

    def test_node(self):
        self.assertIsInstance(
            pubsub_xso.EventDelete.node,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.EventDelete.node.tag,
            (None, "node")
        )
        self.assertIsInstance(
            pubsub_xso.EventDelete.node.type_,
            xso.String
        )
        self.assertIs(
            pubsub_xso.EventDelete.node.default,
            xso.NO_DEFAULT,
        )

    def test_init_default(self):
        with self.assertRaises(TypeError):
            pubsub_xso.EventDelete()

    def test_init(self):
        d = pubsub_xso.EventDelete("node")
        self.assertEqual(d.node, "node")
        self.assertIsNone(d._redirect)

        d = pubsub_xso.EventDelete("bar", redirect_uri="some-uri")
        self.assertEqual(d.node, "bar")
        self.assertIsInstance(d._redirect, pubsub_xso.EventRedirect)
        self.assertEqual(d._redirect.uri, "some-uri")

    def test_redirect_uri_gets_uri_of__redirect(self):
        d = pubsub_xso.EventDelete("foo")
        d._redirect = unittest.mock.Mock()

        self.assertEqual(
            d.redirect_uri,
            d._redirect.uri,
        )

    def test_redirect_uri_returns_none_if_no__redirect_set(self):
        d = pubsub_xso.EventDelete("foo")
        self.assertIsNone(d.redirect_uri)

    def test_setting_redirect_uri_sets__redirect(self):
        d = pubsub_xso.EventDelete("foo")
        d.redirect_uri = "foobar"
        self.assertIsInstance(d._redirect, pubsub_xso.EventRedirect)
        self.assertEqual(
            d._redirect.uri,
            "foobar"
        )

    def test_setting_redirect_uri_to_None_clears__redirect(self):
        d = pubsub_xso.EventDelete("foo")
        d._redirect = unittest.mock.sentinel.foo
        d.redirect_uri = None
        self.assertIsNone(d._redirect)

    def test_deleting_redirect_uri_clears__redirect(self):
        d = pubsub_xso.EventDelete("foo")
        d._redirect = unittest.mock.sentinel.foo
        del d.redirect_uri
        self.assertIsNone(d._redirect)


class TestEventRetract(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.EventRetract,
            xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.EventRetract.TAG,
            (namespaces.xep0060_event, "retract"),
        )

    def test_id_(self):
        self.assertIsInstance(
            pubsub_xso.EventRetract.id_,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.EventRetract.id_.tag,
            (None, "id")
        )
        self.assertIs(
            pubsub_xso.EventRetract.id_.default,
            xso.NO_DEFAULT
        )

    def test_init_default(self):
        with self.assertRaises(TypeError):
            pubsub_xso.EventRetract()

    def test_init(self):
        r = pubsub_xso.EventRetract("some-id")
        self.assertEqual(
            r.id_,
            "some-id",
        )


class TestEventItem(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.EventItem,
            xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.EventItem.TAG,
            (namespaces.xep0060_event, "item"),
        )

    def test_id_(self):
        self.assertIsInstance(
            pubsub_xso.EventItem.id_,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.EventItem.id_.tag,
            (None, "id")
        )
        self.assertIs(
            pubsub_xso.EventItem.id_.default,
            None
        )

    def test_node(self):
        self.assertIsInstance(
            pubsub_xso.EventItem.node,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.EventItem.node.tag,
            (None, "node")
        )
        self.assertIsInstance(
            pubsub_xso.EventItem.node.type_,
            xso.String
        )
        self.assertIs(
            pubsub_xso.EventItem.node.default,
            None,
        )

    def test_publisher(self):
        self.assertIsInstance(
            pubsub_xso.EventItem.publisher,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.EventItem.publisher.tag,
            (None, "publisher")
        )
        self.assertIsInstance(
            pubsub_xso.EventItem.publisher.type_,
            xso.String
        )
        self.assertIs(
            pubsub_xso.EventItem.publisher.default,
            None,
        )

    def test_registered_payload(self):
        self.assertIsInstance(
            pubsub_xso.EventItem.registered_payload,
            xso.Child
        )

    def test_unknown_payload(self):
        self.assertIsInstance(
            pubsub_xso.EventItem.unregistered_payload,
            xso.Collector
        )

    def test_init(self):
        with self.assertRaises(TypeError):
            pubsub_xso.EventItem()

        obj = SomePayload()

        item = pubsub_xso.EventItem(obj)
        self.assertIs(
            item.registered_payload,
            obj
        )

        item = pubsub_xso.EventItem(obj, id_="foo")
        self.assertIs(
            item.registered_payload,
            obj
        )
        self.assertEqual(item.id_, "foo")


class TestEventItems(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.EventItems,
            xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.EventItems.TAG,
            (namespaces.xep0060_event, "items"),
        )

    def test_node(self):
        self.assertIsInstance(
            pubsub_xso.EventItems.node,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.EventItems.node.tag,
            (None, "node")
        )
        self.assertIsInstance(
            pubsub_xso.EventItems.node.type_,
            xso.String
        )
        self.assertIs(
            pubsub_xso.EventItems.node.default,
            xso.NO_DEFAULT,
        )

    def test_retracts(self):
        self.assertIsInstance(
            pubsub_xso.EventItems.retracts,
            xso.ChildList,
        )
        self.assertSetEqual(
            pubsub_xso.EventItems.retracts._classes,
            {
                pubsub_xso.EventRetract,
            }
        )

    def test_items(self):
        self.assertIsInstance(
            pubsub_xso.EventItems.items,
            xso.ChildList,
        )
        self.assertSetEqual(
            pubsub_xso.EventItems.items._classes,
            {
                pubsub_xso.EventItem,
            }
        )

    def test_init_default(self):
        with self.assertRaisesRegex(
                TypeError, "required positional argument: 'node'"):
            pubsub_xso.EventItems()

    def test_init(self):
        items_list = [
            unittest.mock.sentinel.item1,
            unittest.mock.sentinel.item2,
        ]

        retracts_list = [
            unittest.mock.sentinel.retract1,
            unittest.mock.sentinel.retract2,
        ]

        items = pubsub_xso.EventItems(
            items=items_list,
            retracts=retracts_list,
            node="foo"
        )

        self.assertSequenceEqual(
            items.items,
            items_list,
        )

        self.assertSequenceEqual(
            items.retracts,
            retracts_list,
        )

        self.assertEqual(items.node, "foo")

        self.assertIsNot(items.items, items_list)
        self.assertIsNot(items.retracts, retracts_list)


class TestEventPurge(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.EventPurge,
            xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.EventPurge.TAG,
            (namespaces.xep0060_event, "purge"),
        )

    def test_node(self):
        self.assertIsInstance(
            pubsub_xso.EventPurge.node,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.EventPurge.node.tag,
            (None, "node")
        )
        self.assertIsInstance(
            pubsub_xso.EventPurge.node.type_,
            xso.String
        )
        self.assertIs(
            pubsub_xso.EventPurge.node.default,
            xso.NO_DEFAULT,
        )


class TestEventSubscription(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.EventSubscription,
            xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.EventSubscription.TAG,
            (namespaces.xep0060_event, "subscription"),
        )

    def test_jid(self):
        self.assertIsInstance(
            pubsub_xso.EventSubscription.jid,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.EventSubscription.jid.tag,
            (None, "jid")
        )
        self.assertIsInstance(
            pubsub_xso.EventSubscription.jid.type_,
            xso.JID
        )
        self.assertIs(
            pubsub_xso.EventSubscription.jid.default,
            xso.NO_DEFAULT
        )

    def test_node(self):
        self.assertIsInstance(
            pubsub_xso.EventSubscription.node,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.EventSubscription.node.tag,
            (None, "node")
        )
        self.assertIsInstance(
            pubsub_xso.EventSubscription.node.type_,
            xso.String
        )
        self.assertIs(
            pubsub_xso.EventSubscription.node.default,
            None
        )

    def test_subid(self):
        self.assertIsInstance(
            pubsub_xso.EventSubscription.subid,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.EventSubscription.subid.tag,
            (None, "subid")
        )
        self.assertIsInstance(
            pubsub_xso.EventSubscription.subid.type_,
            xso.String
        )
        self.assertIs(
            pubsub_xso.EventSubscription.subid.default,
            None
        )

    def test_subscription(self):
        self.assertIsInstance(
            pubsub_xso.EventSubscription.subscription,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.EventSubscription.subscription.tag,
            (None, "subscription")
        )
        self.assertIsInstance(
            pubsub_xso.EventSubscription.subscription.type_,
            xso.String
        )
        self.assertIsInstance(
            pubsub_xso.EventSubscription.subscription.validator,
            xso.RestrictToSet
        )
        self.assertSetEqual(
            pubsub_xso.EventSubscription.subscription.validator.values,
            {
                "none",
                "pending",
                "subscribed",
                "unconfigured"
            }
        )
        self.assertIs(
            pubsub_xso.EventSubscription.subscription.default,
            None
        )

    def test_expiry(self):
        self.assertIsInstance(
            pubsub_xso.EventSubscription.expiry,
            xso.Attr,
        )
        self.assertEqual(
            pubsub_xso.EventSubscription.expiry.tag,
            (None, "expiry"),
        )
        self.assertIsInstance(
            pubsub_xso.EventSubscription.expiry.type_,
            xso.DateTime,
        )


class TestEventConfiguration(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.EventConfiguration,
            xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.EventConfiguration.TAG,
            (namespaces.xep0060_event, "configuration"),
        )

    def test_node(self):
        self.assertIsInstance(
            pubsub_xso.EventConfiguration.node,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.EventConfiguration.node.tag,
            (None, "node")
        )
        self.assertIsInstance(
            pubsub_xso.EventConfiguration.node.type_,
            xso.String
        )
        self.assertIs(
            pubsub_xso.EventConfiguration.node.default,
            None
        )

    def test_data(self):
        self.assertIsInstance(
            pubsub_xso.EventConfiguration.data,
            xso.Child,
        )
        self.assertSetEqual(
            pubsub_xso.EventConfiguration.data._classes,
            {
                forms.Data,
            }
        )


class TestEvent(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.Event,
            xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.Event.TAG,
            (namespaces.xep0060_event, "event"),
        )

    def test_payload(self):
        self.assertIsInstance(
            pubsub_xso.Event.payload,
            xso.Child,
        )
        self.assertSetEqual(
            pubsub_xso.Event.payload._classes,
            {
                pubsub_xso.EventCollection,
                pubsub_xso.EventConfiguration,
                pubsub_xso.EventDelete,
                pubsub_xso.EventItems,
                pubsub_xso.EventPurge,
                pubsub_xso.EventSubscription,
            }
        )

    def test_is_message_payload(self):
        self.assertIsInstance(
            stanza.Message.xep0060_event,
            xso.Child,
        )
        self.assertSetEqual(
            stanza.Message.xep0060_event._classes,
            {
                pubsub_xso.Event,
            }
        )

    def test_init_default(self):
        ev = pubsub_xso.Event()
        self.assertIsNone(ev.payload)

    def test_init(self):
        items = pubsub_xso.EventItems(node="foo")
        ev = pubsub_xso.Event(items)
        self.assertIs(ev.payload, items)


class TestOwnerAffiliation(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.OwnerAffiliation,
            xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.OwnerAffiliation.TAG,
            (namespaces.xep0060_owner, "affiliation"),
        )

    def test_affiliation(self):
        self.assertIsInstance(
            pubsub_xso.OwnerAffiliation.affiliation,
            xso.Attr,
        )
        self.assertEqual(
            pubsub_xso.OwnerAffiliation.affiliation.tag,
            (None, "affiliation"),
        )
        self.assertIs(
            pubsub_xso.OwnerAffiliation.affiliation.default,
            xso.NO_DEFAULT,
        )
        self.assertIsInstance(
            pubsub_xso.OwnerAffiliation.affiliation.validator,
            xso.RestrictToSet,
        )
        self.assertSetEqual(
            pubsub_xso.OwnerAffiliation.affiliation.validator.values,
            {
                "member",
                "none",
                "outcast",
                "owner",
                "publisher",
                "publish-only",
            }
        )
        self.assertEqual(
            pubsub_xso.OwnerAffiliation.affiliation.validate,
            xso.ValidateMode.ALWAYS,
        )

    def test_jid(self):
        self.assertIsInstance(
            pubsub_xso.OwnerAffiliation.jid,
            xso.Attr,
        )
        self.assertEqual(
            pubsub_xso.OwnerAffiliation.jid.tag,
            (None, "jid"),
        )
        self.assertIs(
            pubsub_xso.OwnerAffiliation.jid.default,
            xso.NO_DEFAULT,
        )
        self.assertIsInstance(
            pubsub_xso.OwnerAffiliation.jid.type_,
            xso.JID,
        )

    def test_init(self):
        with self.assertRaises(TypeError):
            a = pubsub_xso.OwnerAffiliation()

        a = pubsub_xso.OwnerAffiliation(TEST_JID, "owner")
        self.assertEqual(a.jid, TEST_JID)
        self.assertEqual(a.affiliation, "owner")


class TestOwnerAffiliations(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.OwnerAffiliations,
            xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.OwnerAffiliations.TAG,
            (namespaces.xep0060_owner, "affiliations"),
        )

    def test_affiliations(self):
        self.assertIsInstance(
            pubsub_xso.OwnerAffiliations.affiliations,
            xso.ChildList,
        )
        self.assertSetEqual(
            pubsub_xso.OwnerAffiliations.affiliations._classes,
            {
                pubsub_xso.OwnerAffiliation,
            }
        )

    def test_node(self):
        self.assertIsInstance(
            pubsub_xso.OwnerAffiliations.node,
            xso.Attr,
        )
        self.assertEqual(
            pubsub_xso.OwnerAffiliations.node.tag,
            (None, "node"),
        )
        self.assertIs(
            pubsub_xso.OwnerAffiliations.node.default,
            xso.NO_DEFAULT,
        )

    def test_init_default(self):
        with self.assertRaises(TypeError):
            pubsub_xso.OwnerAffiliations()

    def test_init(self):
        as_ = pubsub_xso.OwnerAffiliations("foo")
        self.assertEqual(as_.node, "foo")
        self.assertSequenceEqual(as_.affiliations, [])

        as_ = pubsub_xso.OwnerAffiliations(
            "foo",
            affiliations=[
                unittest.mock.sentinel.a1,
                unittest.mock.sentinel.a2,
            ]
        )
        self.assertEqual(as_.node, "foo")
        self.assertSequenceEqual(
            as_.affiliations,
            [
                unittest.mock.sentinel.a1,
                unittest.mock.sentinel.a2,
            ]
        )


class TestOwnerConfigure(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.OwnerConfigure,
            xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.OwnerConfigure.TAG,
            (namespaces.xep0060_owner, "configure"),
        )

    def test_node(self):
        self.assertIsInstance(
            pubsub_xso.OwnerConfigure.node,
            xso.Attr,
        )
        self.assertEqual(
            pubsub_xso.OwnerConfigure.node.tag,
            (None, "node"),
        )
        self.assertIs(
            pubsub_xso.OwnerConfigure.node.default,
            None,
        )

    def test_data(self):
        self.assertIsInstance(
            pubsub_xso.OwnerConfigure.data,
            xso.Child,
        )
        self.assertSetEqual(
            pubsub_xso.OwnerConfigure.data._classes,
            {
                forms.Data,
            }
        )


class TestOwnerDefault(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.OwnerDefault,
            xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.OwnerDefault.TAG,
            (namespaces.xep0060_owner, "default"),
        )

    def test_data(self):
        self.assertIsInstance(
            pubsub_xso.OwnerDefault.data,
            xso.Child,
        )
        self.assertSetEqual(
            pubsub_xso.OwnerDefault.data._classes,
            {
                forms.Data,
            }
        )


class TestOwnerRedirect(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.OwnerRedirect,
            xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.OwnerRedirect.TAG,
            (namespaces.xep0060_owner, "redirect"),
        )

    def test_uri(self):
        self.assertIsInstance(
            pubsub_xso.OwnerRedirect.uri,
            xso.Attr,
        )
        self.assertEqual(
            pubsub_xso.OwnerRedirect.uri.tag,
            (None, "uri"),
        )
        self.assertIs(
            pubsub_xso.OwnerRedirect.uri.default,
            xso.NO_DEFAULT,
        )

    def test_init_default(self):
        with self.assertRaises(TypeError):
            pubsub_xso.OwnerRedirect()

    def test_init(self):
        r = pubsub_xso.OwnerRedirect("uri")
        self.assertEqual(r.uri, "uri")


class TestOwnerDelete(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.OwnerDelete,
            xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.OwnerDelete.TAG,
            (namespaces.xep0060_owner, "delete"),
        )

    def test__redirect(self):
        self.assertIsInstance(
            pubsub_xso.OwnerDelete._redirect,
            xso.Child,
        )
        self.assertSetEqual(
            pubsub_xso.OwnerDelete._redirect._classes,
            {
                pubsub_xso.OwnerRedirect,
            }
        )

    def test_node(self):
        self.assertIsInstance(
            pubsub_xso.OwnerDelete.node,
            xso.Attr,
        )
        self.assertEqual(
            pubsub_xso.OwnerDelete.node.tag,
            (None, "node"),
        )
        self.assertIs(
            pubsub_xso.OwnerDelete.node.default,
            xso.NO_DEFAULT,
        )

    def test_init_default(self):
        with self.assertRaises(TypeError):
            pubsub_xso.OwnerDelete()

    def test_init(self):
        d = pubsub_xso.OwnerDelete("node")
        self.assertEqual(d.node, "node")
        self.assertIsNone(d._redirect)

        d = pubsub_xso.OwnerDelete(
            "other-node",
            redirect_uri="foo",
        )
        self.assertEqual(d.node, "other-node")
        self.assertIsInstance(d._redirect, pubsub_xso.OwnerRedirect)
        self.assertEqual(
            d._redirect.uri,
            "foo"
        )

    def test_redirect_uri_gets_uri_of__redirect(self):
        d = pubsub_xso.OwnerDelete("foo")
        d._redirect = unittest.mock.Mock()

        self.assertEqual(
            d.redirect_uri,
            d._redirect.uri,
        )

    def test_redirect_uri_returns_none_if_no__redirect_set(self):
        d = pubsub_xso.OwnerDelete("foo")
        self.assertIsNone(d.redirect_uri)

    def test_setting_redirect_uri_sets__redirect(self):
        d = pubsub_xso.OwnerDelete("foo")
        d.redirect_uri = "foobar"
        self.assertIsInstance(d._redirect, pubsub_xso.OwnerRedirect)
        self.assertEqual(
            d._redirect.uri,
            "foobar"
        )

    def test_setting_redirect_uri_to_None_clears__redirect(self):
        d = pubsub_xso.OwnerDelete("foo")
        d._redirect = unittest.mock.sentinel.foo
        d.redirect_uri = None
        self.assertIsNone(d._redirect)

    def test_deleting_redirect_uri_clears__redirect(self):
        d = pubsub_xso.OwnerDelete("foo")
        d._redirect = unittest.mock.sentinel.foo
        del d.redirect_uri
        self.assertIsNone(d._redirect)


class TestOwnerPurge(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.OwnerPurge,
            xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.OwnerPurge.TAG,
            (namespaces.xep0060_owner, "purge"),
        )

    def test_node(self):
        self.assertIsInstance(
            pubsub_xso.OwnerPurge.node,
            xso.Attr,
        )
        self.assertEqual(
            pubsub_xso.OwnerPurge.node.tag,
            (None, "node"),
        )
        self.assertIs(
            pubsub_xso.OwnerPurge.node.default,
            xso.NO_DEFAULT,
        )

    def test_init_default(self):
        with self.assertRaises(TypeError):
            pubsub_xso.OwnerPurge()

    def test_init(self):
        p = pubsub_xso.OwnerDelete("node")
        self.assertEqual(p.node, "node")


class TestOwnerSubscription(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.OwnerSubscription,
            xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.OwnerSubscription.TAG,
            (namespaces.xep0060_owner, "subscription"),
        )

    def test_subscription(self):
        self.assertIsInstance(
            pubsub_xso.OwnerSubscription.subscription,
            xso.Attr,
        )
        self.assertEqual(
            pubsub_xso.OwnerSubscription.subscription.tag,
            (None, "subscription"),
        )
        self.assertIs(
            pubsub_xso.OwnerSubscription.subscription.default,
            xso.NO_DEFAULT,
        )
        self.assertIsInstance(
            pubsub_xso.OwnerSubscription.subscription.validator,
            xso.RestrictToSet
        )
        self.assertSetEqual(
            pubsub_xso.OwnerSubscription.subscription.validator.values,
            {
                "none",
                "pending",
                "subscribed",
                "unconfigured",
            }
        )
        self.assertEqual(
            pubsub_xso.OwnerSubscription.subscription.validate,
            xso.ValidateMode.ALWAYS,
        )

    def test_jid(self):
        self.assertIsInstance(
            pubsub_xso.OwnerSubscription.jid,
            xso.Attr,
        )
        self.assertEqual(
            pubsub_xso.OwnerSubscription.jid.tag,
            (None, "jid"),
        )
        self.assertIs(
            pubsub_xso.OwnerSubscription.jid.default,
            xso.NO_DEFAULT,
        )
        self.assertIsInstance(
            pubsub_xso.OwnerSubscription.jid.type_,
            xso.JID
        )

    def test_init_default(self):
        with self.assertRaises(TypeError):
            pubsub_xso.OwnerSubscription()

    def test_init(self):
        s = pubsub_xso.OwnerSubscription(
            TEST_JID,
            "subscribed"
        )

        self.assertEqual(s.jid, TEST_JID)
        self.assertEqual(s.subscription, "subscribed")


class TestOwnerSubscriptions(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.OwnerSubscriptions,
            xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.OwnerSubscriptions.TAG,
            (namespaces.xep0060_owner, "subscriptions"),
        )

    def test_node(self):
        self.assertIsInstance(
            pubsub_xso.OwnerSubscriptions.node,
            xso.Attr,
        )
        self.assertEqual(
            pubsub_xso.OwnerSubscriptions.node.tag,
            (None, "node"),
        )
        self.assertIs(
            pubsub_xso.OwnerSubscriptions.node.default,
            xso.NO_DEFAULT,
        )

    def test_subscriptions(self):
        self.assertIsInstance(
            pubsub_xso.OwnerSubscriptions.subscriptions,
            xso.ChildList,
        )
        self.assertSetEqual(
            pubsub_xso.OwnerSubscriptions.subscriptions._classes,
            {
                pubsub_xso.OwnerSubscription,
            }
        )

    def test_init_default(self):
        with self.assertRaises(TypeError):
            pubsub_xso.OwnerSubscriptions()

    def test_init(self):
        ss = pubsub_xso.OwnerSubscriptions("node")
        self.assertEqual(ss.node, "node")
        self.assertSequenceEqual(ss.subscriptions, [])

        ss = pubsub_xso.OwnerSubscriptions(
            "node",
            subscriptions=[
                unittest.mock.sentinel.s1,
                unittest.mock.sentinel.s2,
            ]
        )
        self.assertEqual(ss.node, "node")
        self.assertSequenceEqual(
            ss.subscriptions,
            [
                unittest.mock.sentinel.s1,
                unittest.mock.sentinel.s2,
            ]
        )


class TestOwnerRequest(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.OwnerRequest,
            xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(
            pubsub_xso.OwnerRequest.TAG,
            (namespaces.xep0060_owner, "pubsub"),
        )

    def test_payload(self):
        self.assertIsInstance(
            pubsub_xso.OwnerRequest.payload,
            xso.Child,
        )
        self.assertSetEqual(
            pubsub_xso.OwnerRequest.payload._classes,
            {
                pubsub_xso.OwnerAffiliations,
                pubsub_xso.OwnerConfigure,
                pubsub_xso.OwnerDefault,
                pubsub_xso.OwnerDelete,
                pubsub_xso.OwnerPurge,
                pubsub_xso.OwnerSubscriptions,
            }
        )

    def test_is_registered_iq_payload(self):
        self.assertIn(
            pubsub_xso.OwnerRequest,
            stanza.IQ.payload._classes
        )

    def test_init_default(self):
        with self.assertRaises(TypeError):
            pubsub_xso.OwnerRequest()

    def test_init(self):
        r = pubsub_xso.OwnerRequest(unittest.mock.sentinel.payload)
        self.assertEqual(
            r.payload,
            unittest.mock.sentinel.payload
        )


class Testas_payload_class(unittest.TestCase):
    def test_registers_at_EventItem_and_Item(self):
        with contextlib.ExitStack() as stack:
            at_Item = stack.enter_context(
                unittest.mock.patch.object(
                    pubsub_xso.Item,
                    "register_child"
                )
            )

            at_EventItem = stack.enter_context(
                unittest.mock.patch.object(
                    pubsub_xso.EventItem,
                    "register_child"
                )
            )

            result = pubsub_xso.as_payload_class(
                unittest.mock.sentinel.cls
            )

        self.assertIs(result, unittest.mock.sentinel.cls)

        at_Item.assert_called_with(
            pubsub_xso.Item.registered_payload,
            unittest.mock.sentinel.cls,
        )

        at_EventItem.assert_called_with(
            pubsub_xso.EventItem.registered_payload,
            unittest.mock.sentinel.cls,
        )


class TestSimpleErrors(unittest.TestCase):
    ERROR_CLASSES = [
        ("ClosedNode", "closed-node"),
        ("ConfigurationRequired", "configuration-required"),
        ("InvalidJID", "invalid-jid"),
        ("InvalidOptions", "invalid-options"),
        ("InvalidPayload", "invalid-payload"),
        ("InvalidSubID", "invalid-subid"),
        ("ItemForbidden", "item-forbidden"),
        ("ItemRequired", "item-required"),
        ("JIDRequired", "jid-required"),
        ("MaxItemsExceeded", "max-items-exceeded"),
        ("MaxNodesExceeded", "max-nodes-exceeded"),
        ("NodeIDRequired", "nodeid-required"),
        ("NotInRosterGroup", "not-in-roster-group"),
        ("NotSubscribed", "not-subscribed"),
        ("PayloadTooBig", "payload-too-big"),
        ("PayloadRequired", "payload-required"),
        ("PendingSubscription", "pending-subscription"),
        ("PresenceSubscriptionRequired",
         "presence-subscription-required"),
        ("SubIDRequired", "subid-required"),
        ("TooManySubscriptions", "too-many-subscriptions"),
    ]

    def _run_tests(self, func):
        for clsname, *args in self.ERROR_CLASSES:
            cls = getattr(pubsub_xso, clsname)
            func(cls, args)

    def _test_is_xso(self, cls, args):
        self.assertTrue(issubclass(
            cls,
            xso.XSO
        ))

    def test_is_xso(self):
        self._run_tests(self._test_is_xso)

    def _test_tag(self, cls, args):
        self.assertEqual(
            (namespaces.xep0060_errors, args[0]),
            cls.TAG
        )

    def test_tag(self):
        self._run_tests(self._test_tag)

    def _test_is_application_error(self, cls, args):
        self.assertIn(
            cls,
            stanza.Error.application_condition._classes
        )

    def test_is_application_error(self):
        self._run_tests(self._test_is_application_error)


class TestUnsupported(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            pubsub_xso.Unsupported,
            xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            (namespaces.xep0060_errors, "unsupported"),
            pubsub_xso.Unsupported.TAG
        )

    def test_feature(self):
        self.assertIsInstance(
            pubsub_xso.Unsupported.feature,
            xso.Attr
        )
        self.assertEqual(
            pubsub_xso.Unsupported.feature.tag,
            (None, "feature")
        )
        self.assertIsInstance(
            pubsub_xso.Unsupported.feature.validator,
            xso.RestrictToSet
        )
        self.assertSetEqual(
            pubsub_xso.Unsupported.feature.validator.values,
            {
                "access-authorize",
                "access-open",
                "access-presence",
                "access-roster",
                "access-whitelist",
                "auto-create",
                "auto-subscribe",
                "collections",
                "config-node",
                "create-and-configure",
                "create-nodes",
                "delete-items",
                "delete-nodes",
                "filtered-notifications",
                "get-pending",
                "instant-nodes",
                "item-ids",
                "last-published",
                "leased-subscription",
                "manage-subscriptions",
                "member-affiliation",
                "meta-data",
                "modify-affiliations",
                "multi-collection",
                "multi-subscribe",
                "outcast-affiliation",
                "persistent-items",
                "presence-notifications",
                "presence-subscribe",
                "publish",
                "publish-options",
                "publish-only-affiliation",
                "publisher-affiliation",
                "purge-nodes",
                "retract-items",
                "retrieve-affiliations",
                "retrieve-default",
                "retrieve-items",
                "retrieve-subscriptions",
                "subscribe",
                "subscription-options",
                "subscription-notifications",
            }
        )
        self.assertIs(
            pubsub_xso.Unsupported.feature.default,
            xso.NO_DEFAULT
        )

    def test_is_application_error(self):
        self.assertIn(
            pubsub_xso.Unsupported,
            stanza.Error.application_condition._classes
        )


# foo



########################################################################
# File name: test_tasks.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import contextlib
import unittest
import unittest.mock

import aioxmpp.tasks as tasks

from aioxmpp.testutils import CoroutineMock


@asyncio.coroutine
def _infinite_loop():
    while True:
        yield from asyncio.sleep(1)


class TestTaskPool(unittest.TestCase):
    def setUp(self):
        self.p = tasks.TaskPool()

    def tearDown(self):
        del self.p

    def test_defaults(self):
        self.assertIsNone(self.p.get_limit(()))
        self.assertIsNone(self.p.default_limit)

    def test_max_tasks_sets_limit_on_empty_tuple(self):
        p = tasks.TaskPool(max_tasks=10)
        self.assertEqual(
            p.get_limit(()),
            10,
        )

    def test_set_limit(self):
        self.p.set_limit(("foo", "bar"), 2)
        self.assertEqual(self.p.get_limit(("foo", "bar")), 2)
        self.assertIsNone(self.p.get_limit(()))
        self.assertIsNone(self.p.get_limit(("fnord",)))
        self.assertIsNone(self.p.get_limit(("foo",)))

    def test_set_limit_is_idempotent(self):
        self.p.set_limit(("foo", "bar"), 2)
        self.p.set_limit(("foo", "bar"), 2)
        self.assertEqual(self.p.get_limit(("foo", "bar")), 2)

    def test_clear_limit(self):
        self.p.set_limit(("foo", "bar"), 2)
        self.p.clear_limit(("foo", "bar"))
        self.assertIsNone(self.p.get_limit(("foo", "bar")))

    def test_clear_limit_is_idempotent(self):
        self.p.set_limit(("foo", "bar"), 2)
        self.p.clear_limit(("foo", "bar"))
        self.p.clear_limit(("foo", "bar"))
        self.assertIsNone(self.p.get_limit(("foo", "bar")))

    def test_set_limit_with_none_behaves_like_clear_limit(self):
        self.p.set_limit(("foo", "bar"), 2)
        self.p.set_limit(("foo", "bar"), None)
        self.assertIsNone(self.p.get_limit(("foo", "bar")))
        self.p.clear_limit(("foo", "bar"))
        self.assertIsNone(self.p.get_limit(("foo", "bar")))

    def test_empty_groups_have_zero_tasks(self):
        self.assertEqual(self.p.get_task_count(()), 0)
        self.assertEqual(self.p.get_task_count(("foo",)), 0)
        self.assertEqual(self.p.get_task_count(("foo", "bar")), 0)

    def test_spawn_starts_coroutine_and_returns_task(self):
        coro_fun = unittest.mock.Mock()

        with contextlib.ExitStack() as stack:
            async_ = stack.enter_context(
                unittest.mock.patch(
                    "asyncio.ensure_future"
                )
            )

            result = self.p.spawn(
                set(), coro_fun, 1, 2, groups="foo", coro_fun="bar"
            )

        coro_fun.assert_called_once_with(
            1, 2,
            groups="foo",
            coro_fun="bar",
        )

        async_.assert_called_once_with(
            coro_fun()
        )

        self.assertEqual(
            result,
            async_()
        )

    def test_spawn_accounting(self):
        coro_fun = unittest.mock.Mock()

        with contextlib.ExitStack() as stack:
            async_ = stack.enter_context(
                unittest.mock.patch(
                    "asyncio.ensure_future"
                )
            )

            result = self.p.spawn(
                set(), coro_fun, 1, 2, groups="foo", coro_fun="bar"
            )

        coro_fun.assert_called_once_with(
            1, 2,
            groups="foo",
            coro_fun="bar",
        )

        async_.assert_called_once_with(
            coro_fun()
        )

        self.assertEqual(
            result,
            async_()
        )



########################################################################
# File name: test_service.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import types
import unittest

import aioxmpp
import aioxmpp.presence.service as presence_service
import aioxmpp.service as service
import aioxmpp.stanza as stanza
import aioxmpp.structs as structs

from aioxmpp.testutils import (
    make_connected_client,
    run_coroutine,
    CoroutineMock,
)


TEST_PEER_JID1 = structs.JID.fromstr("bar@b.example")
TEST_PEER_JID2 = structs.JID.fromstr("baz@c.example")


class TestPresenceClient(unittest.TestCase):
    def test_is_service(self):
        self.assertTrue(issubclass(
            presence_service.PresenceClient,
            service.Service
        ))

    def setUp(self):
        self.cc = make_connected_client()
        self.presence_dispatcher = aioxmpp.dispatcher.SimplePresenceDispatcher(
            self.cc,
        )
        self.s = presence_service.PresenceClient(self.cc, dependencies={
            aioxmpp.dispatcher.SimplePresenceDispatcher:
                self.presence_dispatcher,
        })

    def test_handle_presence_decorated(self):
        self.assertTrue(
            aioxmpp.dispatcher.is_presence_handler(
                structs.PresenceType.AVAILABLE,
                None,
                presence_service.PresenceClient.handle_presence,
            ),
        )

        self.assertTrue(
            aioxmpp.dispatcher.is_presence_handler(
                structs.PresenceType.UNAVAILABLE,
                None,
                presence_service.PresenceClient.handle_presence,
            ),
        )

        self.assertTrue(
            aioxmpp.dispatcher.is_presence_handler(
                structs.PresenceType.ERROR,
                None,
                presence_service.PresenceClient.handle_presence,
            ),
        )

    def test_return_empty_resource_set_for_arbitrary_jid(self):
        self.assertDictEqual(
            {},
            self.s.get_peer_resources(TEST_PEER_JID1)
        )

    def test_track_available_resources(self):
        st1 = stanza.Presence(type_=structs.PresenceType.AVAILABLE,
                              from_=TEST_PEER_JID1.replace(resource="foo"))
        self.s.handle_presence(st1)

        self.assertDictEqual(
            {
                "foo": st1
            },
            self.s.get_peer_resources(TEST_PEER_JID1)
        )

        st2 = stanza.Presence(type_=structs.PresenceType.AVAILABLE,
                              from_=TEST_PEER_JID1.replace(resource="bar"))
        self.s.handle_presence(st2)

        self.assertDictEqual(
            {
                "foo": st1,
                "bar": st2,
            },
            self.s.get_peer_resources(TEST_PEER_JID1)
        )

        st = stanza.Presence(type_=structs.PresenceType.UNAVAILABLE,
                             from_=TEST_PEER_JID1.replace(resource="foo"))
        self.s.handle_presence(st)

        self.assertDictEqual(
            {
                "bar": st2
            },
            self.s.get_peer_resources(TEST_PEER_JID1)
        )

    def test_get_stanza_returns_None_for_arbitrary_jid(self):
        self.assertIsNone(self.s.get_stanza(
            TEST_PEER_JID1.replace(resource="foo")
        ))

    def test_get_stanza_returns_original_stanza_as_received(self):
        st = stanza.Presence(type_=structs.PresenceType.AVAILABLE,
                             from_=TEST_PEER_JID1.replace(resource="foo"))
        self.s.handle_presence(st)

        self.assertIs(self.s.get_stanza(st.from_), st)

        st = stanza.Presence(type_=structs.PresenceType.AVAILABLE,
                             from_=TEST_PEER_JID1.replace(resource="bar"))
        self.s.handle_presence(st)

        self.assertIs(self.s.get_stanza(st.from_), st)

    def test_get_stanza_returns_error_stanza_for_bare_jid_as_received(self):
        st = stanza.Presence(type_=structs.PresenceType.ERROR,
                             from_=TEST_PEER_JID1)
        self.s.handle_presence(st)

        self.assertIs(self.s.get_stanza(st.from_), st)

        self.assertDictEqual(
            {},
            self.s.get_peer_resources(st.from_.bare())
        )

    def test_get_stanza_returns_error_stanza_for_full_jid_as_received_for_bare_jid(
            self):
        st = stanza.Presence(type_=structs.PresenceType.ERROR,
                             from_=TEST_PEER_JID1)
        self.s.handle_presence(st)

        self.assertIs(self.s.get_stanza(st.from_.replace(resource="foo")), st)

    def test_error_stanza_overrides_all_other_stanzas(self):
        st = stanza.Presence(type_=structs.PresenceType.AVAILABLE,
                             from_=TEST_PEER_JID1.replace(resource="foo"))
        self.s.handle_presence(st)

        self.assertIs(self.s.get_stanza(st.from_), st)

        st = stanza.Presence(type_=structs.PresenceType.ERROR,
                             from_=TEST_PEER_JID1)
        self.s.handle_presence(st)

        self.assertIs(self.s.get_stanza(st.from_.replace(resource="foo")), st)

    def test_get_any_non_error_stanza_erases_error_stanza(self):
        st = stanza.Presence(type_=structs.PresenceType.ERROR,
                             from_=TEST_PEER_JID1)
        self.s.handle_presence(st)

        self.assertIs(self.s.get_stanza(st.from_), st)

        st = stanza.Presence(type_=structs.PresenceType.UNAVAILABLE,
                             from_=TEST_PEER_JID1.replace(resource="foo"))
        self.s.handle_presence(st)

        self.assertIsNone(self.s.get_stanza(st.from_.bare()))
        self.assertIsNone(self.s.get_stanza(st.from_))

        st = stanza.Presence(type_=structs.PresenceType.AVAILABLE,
                             from_=TEST_PEER_JID1.replace(resource="foo"))
        self.s.handle_presence(st)

        self.assertIs(self.s.get_stanza(st.from_), st)

    def test_get_most_available_stanza(self):
        st = stanza.Presence(type_=structs.PresenceType.AVAILABLE,
                             from_=TEST_PEER_JID1.replace(resource="foo"))
        self.s.handle_presence(st)

        self.assertIs(
            self.s.get_most_available_stanza(TEST_PEER_JID1),
            st
        )

        staway = stanza.Presence(type_=structs.PresenceType.AVAILABLE,
                                 show=structs.PresenceShow.AWAY,
                                 from_=TEST_PEER_JID1.replace(resource="baz"))
        self.s.handle_presence(staway)

        self.assertIs(
            self.s.get_most_available_stanza(TEST_PEER_JID1),
            st
        )

        stdnd = stanza.Presence(type_=structs.PresenceType.AVAILABLE,
                                show=structs.PresenceShow.DND,
                                from_=TEST_PEER_JID1.replace(resource="bar"))
        self.s.handle_presence(stdnd)

        self.assertEqual(
            len(self.s.get_peer_resources(TEST_PEER_JID1)),
            3
        )

        self.assertIs(
            self.s.get_most_available_stanza(TEST_PEER_JID1),
            stdnd
        )

    def test_get_most_available_stanza_returns_None_for_unavailable_JID(self):
        self.assertIsNone(self.s.get_most_available_stanza(TEST_PEER_JID1))

    def test_handle_presence_emits_available_signals(self):
        base = unittest.mock.Mock()
        base.bare.return_value = False
        base.full.return_value = False

        self.s.on_bare_available.connect(base.bare)
        self.s.on_available.connect(base.full)

        st1 = stanza.Presence(type_=structs.PresenceType.AVAILABLE,
                              from_=TEST_PEER_JID1.replace(resource="foo"))
        self.s.handle_presence(st1)

        st2 = stanza.Presence(type_=structs.PresenceType.AVAILABLE,
                              from_=TEST_PEER_JID1.replace(resource="bar"))
        self.s.handle_presence(st2)

        self.assertSequenceEqual(
            base.mock_calls,
            [
                unittest.mock.call.bare(st1),
                unittest.mock.call.full(st1.from_, st1),
                unittest.mock.call.full(st2.from_, st2),
            ]
        )

    def test_handle_presence_emits_available_signals_only_if_not_available(self):
        base = unittest.mock.Mock()
        base.bare.return_value = False
        base.full.return_value = False

        self.s.on_bare_available.connect(base.bare)
        self.s.on_available.connect(base.full)

        st1 = stanza.Presence(type_=structs.PresenceType.AVAILABLE,
                              from_=TEST_PEER_JID1.replace(resource="foo"))
        self.s.handle_presence(st1)

        st2 = stanza.Presence(type_=structs.PresenceType.AVAILABLE,
                              from_=TEST_PEER_JID1.replace(resource="bar"))
        self.s.handle_presence(st2)

        st3 = stanza.Presence(type_=structs.PresenceType.AVAILABLE,
                              from_=TEST_PEER_JID1.replace(resource="bar"))
        self.s.handle_presence(st3)

        self.assertSequenceEqual(
            base.mock_calls,
            [
                unittest.mock.call.bare(st1),
                unittest.mock.call.full(st1.from_, st1),
                unittest.mock.call.full(st2.from_, st2),
            ]
        )

    def test_handle_presence_does_not_emit_unavailable_if_already_unavailable(self):
        base = unittest.mock.Mock()
        base.bare.return_value = False
        base.full.return_value = False

        self.s.on_bare_unavailable.connect(base.bare)
        self.s.on_unavailable.connect(base.full)

        st2 = stanza.Presence(type_=structs.PresenceType.UNAVAILABLE,
                              from_=TEST_PEER_JID1.replace(resource="bar"))
        self.s.handle_presence(st2)

        st1 = stanza.Presence(type_=structs.PresenceType.UNAVAILABLE,
                              from_=TEST_PEER_JID1.replace(resource="foo"))
        self.s.handle_presence(st1)

        self.assertSequenceEqual(
            base.mock_calls,
            [
            ]
        )


    def test_handle_presence_emits_unavailable_signals(self):
        base = unittest.mock.Mock()
        base.bare.return_value = False
        base.full.return_value = False

        self.s.on_bare_unavailable.connect(base.bare)
        self.s.on_unavailable.connect(base.full)

        self.s.handle_presence(
            stanza.Presence(type_=structs.PresenceType.AVAILABLE,
                            from_=TEST_PEER_JID1.replace(resource="foo"))
        )

        self.s.handle_presence(
            stanza.Presence(type_=structs.PresenceType.AVAILABLE,
                            from_=TEST_PEER_JID1.replace(resource="bar"))
        )

        st2 = stanza.Presence(type_=structs.PresenceType.UNAVAILABLE,
                              from_=TEST_PEER_JID1.replace(resource="bar"))
        self.s.handle_presence(st2)

        st1 = stanza.Presence(type_=structs.PresenceType.UNAVAILABLE,
                              from_=TEST_PEER_JID1.replace(resource="foo"))
        self.s.handle_presence(st1)

        self.assertSequenceEqual(
            base.mock_calls,
            [
                unittest.mock.call.full(st2.from_, st2),
                unittest.mock.call.full(st1.from_, st1),
                unittest.mock.call.bare(st1),
            ]
        )

    def test_handle_presence_emits_changed_signals(self):
        base = unittest.mock.Mock()
        base.bare.return_value = False
        base.full.return_value = False

        self.s.on_changed.connect(base.full)

        self.s.handle_presence(
            stanza.Presence(type_=structs.PresenceType.AVAILABLE,
                            from_=TEST_PEER_JID1.replace(resource="foo"))
        )

        self.s.handle_presence(
            stanza.Presence(type_=structs.PresenceType.AVAILABLE,
                            from_=TEST_PEER_JID1.replace(resource="bar"))
        )

        st1 = stanza.Presence(type_=structs.PresenceType.AVAILABLE,
                              show=structs.PresenceShow.DND,
                              from_=TEST_PEER_JID1.replace(resource="foo"))
        self.s.handle_presence(st1)

        st2 = stanza.Presence(type_=structs.PresenceType.AVAILABLE,
                              show=structs.PresenceShow.DND,
                              from_=TEST_PEER_JID1.replace(resource="bar"))
        self.s.handle_presence(st2)

        st3 = stanza.Presence(type_=structs.PresenceType.AVAILABLE,
                              show=structs.PresenceShow.CHAT,
                              from_=TEST_PEER_JID1.replace(resource="bar"))
        self.s.handle_presence(st3)

        self.assertSequenceEqual(
            base.mock_calls,
            [
                unittest.mock.call.full(st1.from_, st1),
                unittest.mock.call.full(st2.from_, st2),
                unittest.mock.call.full(st3.from_, st3),
            ]
        )

    def test_handle_presence_emits_unavailable_on_error(self):
        base = unittest.mock.Mock()
        base.bare.return_value = False
        base.full.return_value = False

        self.s.on_unavailable.connect(base.full)
        self.s.on_bare_unavailable.connect(base.bare)

        self.s.handle_presence(
            stanza.Presence(type_=structs.PresenceType.AVAILABLE,
                            from_=TEST_PEER_JID1.replace(resource="foo"))
        )

        self.s.handle_presence(
            stanza.Presence(type_=structs.PresenceType.AVAILABLE,
                            from_=TEST_PEER_JID1.replace(resource="bar"))
        )

        st = stanza.Presence(type_=structs.PresenceType.ERROR,
                             from_=TEST_PEER_JID1)
        self.s.handle_presence(st)

        self.assertIn(
            unittest.mock.call.full(st.from_.replace(resource="foo"), st),
            base.mock_calls
        )

        self.assertIn(
            unittest.mock.call.full(st.from_.replace(resource="bar"), st),
            base.mock_calls
        )

        self.assertIn(
            unittest.mock.call.bare(st),
            base.mock_calls
        )

    def tearDown(self):
        del self.s
        del self.cc


class TestPresenceServer(unittest.TestCase):
    def setUp(self):
        self.cc = make_connected_client()
        self.cc.send = CoroutineMock()
        self.cc.send.return_value = None
        self.s = presence_service.PresenceServer(self.cc)

    def tearDown(self):
        del self.s
        del self.cc

    def test_no_presence_to_emit_by_default(self):
        self.assertEqual(self.s.state, aioxmpp.PresenceState(False))
        self.assertDictEqual(self.s.status, {})
        self.assertEqual(self.s.priority, 0)

        run_coroutine(self.cc.before_stream_established())

        self.cc.send.assert_not_called()

    def test_before_stream_established_handler_returns_true_for_unav(self):
        self.assertTrue(
            run_coroutine(self.s._before_stream_established())
        )

    def test_before_stream_established_handler_returns_true_for_avail(self):
        self.s.set_presence(aioxmpp.PresenceState(True))

        self.assertTrue(
            run_coroutine(self.s._before_stream_established())
        )

    def test_set_presence_rejects_non_PresenceState_state(self):
        with self.assertRaisesRegex(
                TypeError,
                r"invalid state: got <enum '.*PresenceType'>, "
                r"expected aioxmpp.PresenceState"):
            self.s.set_presence(aioxmpp.PresenceType.AVAILABLE)

    def test_set_presence_with_status_string(self):
        # not established
        self.cc.established = False

        self.s.set_presence(
            aioxmpp.PresenceState(True),
            status="foo"
        )

        self.assertEqual(
            self.s.state,
            aioxmpp.PresenceState(True)
        )

        self.assertDictEqual(
            self.s.status,
            {
                None: "foo",
            }
        )

    def test_set_presence_with_status_mapping(self):
        # not established
        self.cc.established = False

        m = {
            None: "foo",
            aioxmpp.structs.LanguageTag.fromstr("en-gb"): "bar",
        }

        self.s.set_presence(
            aioxmpp.PresenceState(True),
            status=m,
        )

        self.assertEqual(
            self.s.state,
            aioxmpp.PresenceState(True)
        )

        self.assertDictEqual(
            self.s.status,
            m,
        )

        self.assertIsNot(self.s.status, m)

        del m[None]

        self.assertIn(None, self.s.status)

    def test_set_presence_priority(self):
        self.cc.established = False

        self.s.set_presence(aioxmpp.PresenceState(False), priority=10)

        self.assertEqual(
            self.s.priority,
            10,
        )

    def test_set_presence_rejects_non_integer_priority(self):
        self.cc.established = False

        with self.assertRaisesRegex(
                TypeError,
                r"invalid priority: got <class 'str'>, expected integer"):
            self.s.set_presence(
                aioxmpp.PresenceState(True),
                status="foo",
                priority="10",
            )

        self.assertEqual(self.s.state, aioxmpp.PresenceState(False))
        self.assertDictEqual(self.s.status, {})

    def test_emit_configured_presence_when_stream_establishes(self):
        self.s.set_presence(aioxmpp.PresenceState(True))

        with unittest.mock.patch.object(self.s, "make_stanza") as make_stanza:
            make_stanza.return_value = unittest.mock.sentinel.presence
            run_coroutine(self.cc.before_stream_established())

        self.cc.send.assert_called_with(
            unittest.mock.sentinel.presence,
        )

    def test_set_presence_broadcasts_if_established(self):
        self.cc.established = True

        def check_state():
            self.assertEqual(self.cc.state, aioxmpp.PresenceState(True))
            self.assertEqual(self.cc.status, {None: "foo"})

        with unittest.mock.patch.object(self.s, "make_stanza") as make_stanza:
            make_stanza.return_value = unittest.mock.sentinel.presence

            result = self.s.set_presence(
                aioxmpp.PresenceState(True),
                status="foo",
            )

        self.cc.enqueue.assert_called_with(
            unittest.mock.sentinel.presence
        )

        self.assertEqual(
            result,
            self.cc.enqueue()
        )

    def test_make_stanza_converts_state_to_stanza(self):
        m = unittest.mock.Mock(spec=aioxmpp.PresenceState)
        self.s.set_presence(m)

        stanza = self.s.make_stanza()
        self.assertIsInstance(
            stanza,
            aioxmpp.Presence,
        )

        m.apply_to_stanza.assert_called_with(stanza)

    def test_make_stanza_converts_incorporates_status(self):
        self.s.set_presence(aioxmpp.PresenceState(True), status="foo")

        stanza = self.s.make_stanza()
        self.assertDictEqual(
            stanza.status,
            {None: "foo"}
        )

    def test_set_presence_state_emits_events(self):
        new_state = aioxmpp.PresenceState(True)
        new_status = {None: "foo"}
        new_priority = -2

        def check_values():
            self.assertEqual(
                self.s.state,
                new_state,
            )
            self.assertDictEqual(
                self.s.status,
                new_status,
            )
            self.assertEqual(
                self.s.priority,
                new_priority,
            )

        overall_cb = unittest.mock.Mock()
        overall_cb.side_effect = check_values
        state_cb = unittest.mock.Mock()
        state_cb.side_effect = check_values

        self.s.on_presence_changed.connect(overall_cb)
        self.s.on_presence_state_changed.connect(state_cb)

        self.s.set_presence(
            new_state,
            status=new_status,
            priority=new_priority,
        )

        overall_cb.assert_called_once_with()
        state_cb.assert_called_once_with()

    def test_set_presence_state_does_not_emit_state_event_if_state_unchanged(self):
        new_state = aioxmpp.PresenceState(False)
        new_status = {None: "foo"}
        new_priority = -2

        def check_values():
            self.assertEqual(
                self.s.state,
                new_state,
            )
            self.assertDictEqual(
                self.s.status,
                new_status,
            )
            self.assertEqual(
                self.s.priority,
                new_priority,
            )

        overall_cb = unittest.mock.Mock()
        overall_cb.side_effect = check_values
        state_cb = unittest.mock.Mock()
        state_cb.side_effect = check_values

        self.s.on_presence_changed.connect(overall_cb)
        self.s.on_presence_state_changed.connect(state_cb)

        self.s.set_presence(
            new_state,
            status=new_status,
            priority=new_priority,
        )

        overall_cb.assert_called_once_with()
        state_cb.assert_not_called()

    def test_set_presence_state_does_not_emit_events_if_unchanged(self):
        new_state = aioxmpp.PresenceState(False)
        new_status = {}
        new_priority = 0

        def check_values():
            self.assertEqual(
                self.s.state,
                new_state,
            )
            self.assertDictEqual(
                self.s.status,
                new_status,
            )
            self.assertEqual(
                self.s.priority,
                new_priority,
            )

        overall_cb = unittest.mock.Mock()
        overall_cb.side_effect = check_values
        state_cb = unittest.mock.Mock()
        state_cb.side_effect = check_values

        self.s.on_presence_changed.connect(overall_cb)
        self.s.on_presence_state_changed.connect(state_cb)

        self.s.set_presence(
            new_state,
            status=new_status,
            priority=new_priority,
        )

        overall_cb.assert_not_called()
        state_cb.assert_not_called()

    def test_set_presence_state_does_not_emit_stanza_if_unchanged(self):
        new_state = aioxmpp.PresenceState(False)
        new_status = {}
        new_priority = 0

        self.s.set_presence(
            new_state,
            status=new_status,
            priority=new_priority,
        )

        self.cc.enqueue.assert_not_called()

    def test_resend_presence_broadcasts_if_established(self):
        self.cc.established = True

        with unittest.mock.patch.object(self.s, "make_stanza") as make_stanza:
            make_stanza.return_value = unittest.mock.sentinel.presence

            result = self.s.resend_presence()

        self.cc.enqueue.assert_called_with(
            unittest.mock.sentinel.presence
        )

        self.assertEqual(
            result,
            self.cc.enqueue(),
        )
